#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
import logging
import re

import inquirer
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import yaml
from typing_extensions import TypedDict, NotRequired, Self

from quark.modules.Core import Core
from quark.modules.applications import Application
from quark.utils import _get_instance_with_sub_options, checkbox


class ConfigModule(TypedDict):
    """
    Each instance consists of the name of the module, its config, and a list of its configured submodules, which are
    ConfigModule instances themselves. It can also contain the instance of the class associated to this module,
    which is then used in the actual benchmarking process.
    """
    name: str
    config: dict
    submodules: list[Self]
    instance: NotRequired[Core]


class BenchmarkConfig(TypedDict):
    """
    Each instance consists of a ConfigModule associated with the application at the first level and the number of
    repetitions of the benchmark.
    """
    repetitions: int
    application: ConfigModule


class ConfigManager:
    """
    A class responsible for generating/loading QUARK benchmark configs. Loading includes instantiation of the various
    modules specified in the config.
    """

    def __init__(self):
        self.config = None
        self.application = None

    def generate_benchmark_configs(self, app_modules: list[dict]) -> None:
        """
        Queries the user to get all information about the application, other modules, and general settings
        necessary to run the benchmark.

        :param app_modules: List of application modules as specified in the application modules configuration
        """
        application_answer = inquirer.prompt([inquirer.List('application',
                                                            message="What application do you want?",
                                                            choices=[m["name"] for m in app_modules],
                                                            default='PVC',
                                                            )])

        app_name = application_answer["application"]
        self.application = _get_instance_with_sub_options(app_modules, app_name)

        application_config = self.application.get_parameter_options()
        more_app_config = {}

        full_application_config = ConfigManager._query_for_config(
            application_config, f"(Option for {application_answer['application']})")
        depending_submodules = False

        available_submodules = []
        for parameter in application_config:
            if ("depending_submodule" in application_config[parameter] and
                    application_config[parameter]["depending_submodule"]):
                available_submodules = self.application.get_available_submodules(full_application_config[parameter])
                depending_submodules = True
        if not depending_submodules:
            available_submodules = self.application.get_available_submodule_options()
        submodule_answer = checkbox(key='submodules',
                                    message="What submodule do you want?",
                                    choices=available_submodules)
        for option in submodule_answer["submodules"]:
            if self.application.depending_parameters:
                more_app_config = self.application.get_depending_parameters(option, full_application_config)
                more_app_config = (ConfigManager._query_for_config
                                   (more_app_config, f"(Option for {self.application.__class__.__name__})"))
                full_application_config = full_application_config | more_app_config
        self.config = {
            "application": {
                "name": app_name,
                "config": full_application_config,
                "submodules": [self.query_module(self.application.get_submodule(sm), sm) for sm in
                               submodule_answer["submodules"]],
            },
            "repetitions": 1  # default value

        }
        logging.info("Submodule configuration finished")

        repetitions_answer = inquirer.prompt(
            [inquirer.Text('repetitions', message="How many repetitions do you want?",
                           validate=lambda _, x: re.match("\\d", x), default=1)])
        self.config["repetitions"] = int(repetitions_answer["repetitions"])

    def query_module(self, module: Core, module_friendly_name: str) -> ConfigModule:
        """
        Recursive function which queries every module and its submodule until end is reached.

        :param module: Module instance
        :param module_friendly_name: Name of the module
        :return: Config module with the choices of the user
        """
        module_config = module.get_parameter_options()
        full_module_config = ConfigManager._query_for_config(module_config, f"(Option for {module.__class__.__name__})")
        depending_submodules = False
        available_submodules = []
        for parameter in module_config:
            if ("depending_submodule" in module_config[parameter] and
                    module_config[parameter]["depending_submodule"]):
                available_submodules = module.get_available_submodules(full_module_config[parameter])
                depending_submodules = True
        if not depending_submodules:
            available_submodules = module.get_available_submodule_options()
        more_module_config = {}
        if available_submodules:
            if len(available_submodules) == 1:
                logging.info(
                    f"Skipping asking for submodule, since only 1 option ({available_submodules[0]}) is available.")
                submodule_answer = {"submodules": [available_submodules[0]]}
            else:
                submodule_answer = checkbox(key='submodules',
                                            message="What submodule do you want?",
                                            choices=available_submodules)
                for option in submodule_answer:
                    if module.depending_parameters:
                        more_module_config = module.get_depending_parameters(option, full_module_config)
                        more_module_config = (ConfigManager._query_for_config
                                              (more_module_config, f"(Option for {module.__class__.__name__})"))
                        full_module_config = full_module_config | more_module_config
        else:
            submodule_answer = {"submodules": []}

        return {
            "name": module_friendly_name,
            "config": full_module_config,
            "submodules": [self.query_module(module.get_submodule(sm), sm) for sm in
                           submodule_answer["submodules"]]
        }

    def set_config(self, config: BenchmarkConfig) -> None:
        """
        In case the user provides a config file, this function is used to set the config.

        :param config: Valid config file
        """
        if ConfigManager.is_legacy_config(config):
            config = ConfigManager.translate_legacy_config(config)
        self.config = config

    @staticmethod
    def is_legacy_config(config: dict) -> bool:
        """
        Checks if a QUARK 1 config was provided.

        :param config: Valid config file
        :return: True if provided config is QUARK 1 config
        """
        if "mapping" in config.keys():
            logging.warning("Looks like you have provided a legacy config file")
            return True

        return False

    @staticmethod
    def translate_legacy_config(config: dict) -> BenchmarkConfig:
        """
        Translates the QUARK 1 config format to QUARK 2 format.

        :param config: QUARK 1 config
        :return: Translated config
        """
        logging.info("Trying to translate QUARK 1 config to QUARK 2 config format")
        try:
            translated_config = {key: config[key] for key in ["application", "repetitions"]}
            translated_config["application"]["submodules"] = []

            for key, value in config["mapping"].items():
                # Direct mapping does not exist anymore in QUARK 2.0
                if key.lower() == "direct":
                    translated_config["application"]["submodules"].extend(
                        ConfigManager.translate_legacy_config_helper(value, "solver"))
                else:
                    translated_config["application"]["submodules"].append(
                        {
                            "name": key,
                            "config": value["config"],
                            "submodules": ConfigManager.translate_legacy_config_helper(value, "solver")
                        }
                    )
        except Exception as e:
            raise Exception("Error translating the given legacy config file!") from e

        return translated_config

    @staticmethod
    def translate_legacy_config_helper(config_part: dict, module_key: str) -> list:
        """
        Helper function for translate_legacy_config, which translates the QUARK 1 config format to QUARK 2 format.

        :param config_part: Part of a config
        :param module_key: Module key: mapping, solver or device
        :return: Translated config_part
        """
        next_module_key = "device" if module_key.lower() == "solver" else None

        result = []
        for item in config_part[module_key]:
            result.append(
                {
                    "name": item["name"],
                    "config": item["config"],
                    "submodules": ConfigManager.translate_legacy_config_helper(item, next_module_key)
                    if next_module_key is not None else []
                }
            )

        return result

    def load_config(self, app_modules: list[dict]) -> None:
        """
        Uses the config to generate all class instances needed to run the benchmark.

        :param app_modules: List of application modules as specified in the application modules configuration
        """

        self.application = _get_instance_with_sub_options(app_modules, self.config["application"]["name"])
        self.config["application"].update({"instance": self.application,
                                           "submodules": [ConfigManager.initialize_module_classes(self.application, c)
                                                          for c in self.config["application"]["submodules"]]})

    @staticmethod
    def initialize_module_classes(parent_module: Core, config: ConfigModule) -> ConfigModule:
        """
        Recursively initializes all instances of the required modules and their submodules for a given config.

        :param parent_module: Class of the parent module
        :param config: Uninitialized config module
        :return: Config with instances
        """
        module_instance = parent_module.get_submodule(config["name"])
        config.update({"instance": module_instance,
                       "submodules": [ConfigManager.initialize_module_classes(module_instance, c)
                                      for c in config["submodules"]]})
        return config

    def get_config(self) -> BenchmarkConfig:
        """
        Returns the config.

        :return: config
        """
        return self.config

    def get_app(self) -> Application:
        """
        Returns instance of the application.

        :return: Instance of the application
        """
        return self.config["application"]["instance"]

    def get_reps(self) -> int:
        """
        Returns number of repetitions specified in config.

        :return: Number of repetitions
        """
        return self.config["repetitions"]

    def start_create_benchmark_backlog(self) -> list:
        """
        Helper function to kick off the creation of the benchmark backlog.

        :return: List with all benchmark items
        """
        return ConfigManager.create_benchmark_backlog(self.config["application"])

    @staticmethod
    def create_benchmark_backlog(module: ConfigModule) -> list:
        """
        Recursive function which splits up the loaded config into single benchmark runs.

        :param module: ConfigModule
        :return: List with all benchmark items
        """
        items = []

        if len(module["config"].items()) > 0:
            keys, values = zip(*module["config"].items())
            config_expanded = [dict(zip(keys, v)) for v in itertools.product(*values)]
        else:
            config_expanded = [{}]

        for single_config in config_expanded:
            # Check if we reached the end of the chain
            if len(module["submodules"]) > 0:
                for submodule in module["submodules"]:
                    for item in ConfigManager.create_benchmark_backlog(submodule):
                        items.append({
                            "name": module["name"],
                            "instance": module["instance"],
                            "config": single_config,
                            "submodule": {submodule["name"]: item}
                        })
            else:
                items.append({
                    "name": module["name"],
                    "instance": module["instance"],
                    "config": single_config,
                    "submodule": {}
                })

        return items

    def save(self, store_dir: str) -> None:
        """
        Saves the config as a YAML file.

        :param store_dir: Directory in which the file should be stored
        """
        with open(f"{store_dir}/config.yml", 'w') as filehandler:
            yaml.dump(self.config, filehandler)

    def print(self) -> None:
        """
        Prints the config.
        """
        print(yaml.dump(self.config))

    @staticmethod
    def _query_for_config(param_opts: dict, prefix: str = "") -> dict:
        """
        For a given module config, queries users in an interactive mode which of the options they would like to
        include in the final benchmark config.

        :param param_opts: Dictionary containing the options for a parameter including a description
        :param prefix: Prefix string, which is attached when interacting with the user
        :return: Dictionary containing the decisions of the user on what to include in the benchmark
        """
        config = {}
        for key, config_answer in param_opts.items():
            if config_answer.get("if"):
                key_in_cond = config_answer.get("if")["key"]
                dependency = param_opts.get(key_in_cond)
                consistent = False
                err_msg = None
                if dependency is None:
                    err_msg = f"Inconsistent parameter options: condition references unknown parameter: {key_in_cond}"
                elif not dependency.get('exclusive', False):
                    err_msg = f"Inconsistent parameter options: " \
                        f"condition references non exclusive parameter: {key_in_cond}"
                else:
                    consistent = True
                if not consistent:
                    raise Exception(f"{prefix} {err_msg}")

                if config[key_in_cond][0] not in config_answer.get("if")["in"]:
                    continue

            if len(config_answer['values']) == 1 and "custom_input" not in config_answer:
                # When there is only 1 value to choose from skip the user input for now
                values = config_answer['values']
                print(f"{prefix} {config_answer['description']}: {config_answer['values'][0]}")
            elif config_answer.get('exclusive', False):
                answer = inquirer.prompt(
                    [inquirer.List(key, message=f"{prefix} {config_answer['description']}",
                                   choices=config_answer['values'])])
                values = (answer[key],)
            else:
                choices = [*config_answer['values'], "Custom Input"] if (config_answer.get("custom_input") and
                                                                         config_answer["custom_input"]) \
                    else config_answer['values']
                if config_answer.get("allow_ranges") and config_answer["allow_ranges"]:
                    choices.append("Custom Range")

                 # Add custom_input if it is specified in the parameters
                answer = checkbox(key=key, message=f"{prefix} {config_answer['description']}", choices=choices)
                values = answer[key]

                if "Custom Input" in values:
                    freetext_answer = inquirer.prompt(
                        [inquirer.Text('custom_input', message=f"What's your custom input for {key}? (No validation of "
                                       "this input is done!)")])

                    # Replace the freetext placeholder with the user input
                    values.remove("Custom Input")
                    values.append(freetext_answer["custom_input"])

                if "Custom Range" in values:
                    range_answer = inquirer.prompt(
                        [inquirer.Text('start', message=f"What's the start of your range for {key}? (No validation of "
                                       "this input is done!)"),
                         inquirer.Text('stop', message=f"What's the end of your range for {key}? (No validation of "
                                       "this input is done!)"),
                         inquirer.Text('step', message=f"What are the steps of your range for {key}? (No validation of "
                                       "this input is done!)")])
                    values.remove("Custom Range")
                    values.extend(np.arange(float(range_answer["start"]), float(range_answer["stop"]),
                                            float(range_answer["step"])))
                    # Remove possible duplicates
                    values = list(set(values))

            if config_answer.get("postproc"):
                # the value of config_answer.get("postproc") is expected to be callable
                # with each of the user selected values as argument.
                # Note that the stored config file will contain the processed values.
                values = [config_answer["postproc"](v) for v in values]

            config[key] = values

        return config

    def create_tree_figure(self, store_dir: str) -> None:
        """
        Visualizes the benchmark as a graph (experimental feature).

        :param store_dir: Directory where the file should be stored
        """
        graph = nx.DiGraph()
        ConfigManager._create_tree_figure_helper(graph, self.config["application"])
        nx.draw(graph, with_labels=True, pos=nx.spectral_layout(graph), node_shape="s")
        plt.savefig(f"{store_dir}/BenchmarkGraph.png", format="PNG")
        plt.clf()

    @staticmethod
    def _create_tree_figure_helper(graph: nx.Graph, config: ConfigModule) -> None:
        """
        Helper function for create_tree_figure that traverses the config recursively.

        :param graph: Networkx Graph
        :param config: Benchmark config
        """
        if config:
            key = config["name"]
            if "submodules" in config and config["submodules"]:
                for submodule in config["submodules"]:
                    graph.add_edge(key, submodule["name"])
                    ConfigManager._create_tree_figure_helper(graph, submodule)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import re
import logging

import numpy as np
from qiskit_optimization import QuadraticProgram
from qiskit_optimization.converters import (
    InequalityToEquality, IntegerToBinary,
    LinearEqualityToPenalty
)

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement

# TODO Large chunks of this code is duplicated in ACL.mappings.ISING -> unify


class Qubo(Mapping):
    """
    QUBO formulation for the ACL.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]
        self.global_variables = []

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: list of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "qiskit-optimization", "version": "0.6.1"},
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this mapping has no configurable settings.

        :return: Empty dictionary
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def map_pulp_to_qiskit(self, problem: dict) -> QuadraticProgram:
        """
        Maps the problem dict to a quadratic program.

        :param problem: Problem formulation in dict form
        :return: Quadratic program in qiskit-optimization format
        """
        # Details at:
        # https://coin-or.github.io/pulp/guides/how_to_export_models.html
        # https://qiskit.org/documentation/stable/0.26/tutorials/optimization/2_converters_for_quadratic_programs.html

        qp = QuadraticProgram()

        # Variables
        for variable_dict in problem["variables"]:
            if variable_dict["cat"] == "Integer":
                lb = variable_dict["lowBound"]
                ub = variable_dict["upBound"]
                name = variable_dict["name"]
                # If the integer variable is actually a binary variable
                if lb == 0 and ub == 1:
                    qp.binary_var(name)
                # If the integer variable is non-binary
                else:
                    qp.integer_var(lowerbound=lb, upperbound=ub, name=name)

        # Objective function
        obj_arguments = {arg["name"]: arg["value"] for arg in problem["objective"]["coefficients"]}

        # Maximize
        if problem["parameters"]["sense"] == -1:
            qp.maximize(linear=obj_arguments)
        # Minimize
        else:
            qp.minimize(linear=obj_arguments)

        # Constraints
        for constraint in problem["constraints"]:
            const_arguments = {arg["name"]: arg["value"] for arg in constraint["coefficients"]}
            sense = constraint["sense"]
            const_sense = "LE" if sense == -1 else "GE" if sense == 1 else "E"

            qp.linear_constraint(
                linear=const_arguments,
                sense=const_sense,
                rhs=-1 * constraint["constant"],
                name=constraint["name"]
            )

        return qp

    def convert_string_to_arguments(self, input_string: str) -> list[any]:
        """
        Converts QUBO in string format to a list of separated arguments,
        used to construct the QUBO matrix.

        :param input_string: QUBO in raw string format
        :return: List of arguments
        """
        terms = re.findall(r'[+\-]?[^+\-]+', input_string)
        # Convert the penalty string to a list of lists of the individual arguments in the penalty term
        result = [term.strip() for term in terms]
        separated_arguments = []
        first_item = True

        for argument in result:
            if first_item:
                # Remove "maximize" or minimize string from the first argument
                argument = argument[8:]
                first_item = False
            if "*" in argument:
                # The variables in each argument are connected by "*" signs. Here we split the variables
                elements = argument.split('*')
                # Convert string of numbers to floats
                new_argument = elements[0].strip()
                # Remove empty strings
                new_argument = [int(new_argument.replace(" ", "")) if new_argument.replace(" ", "").isdigit()
                                else float(new_argument.replace(" ", ""))]
                new_argument += [el.strip() for el in elements[1:]]
                separated_arguments.append(new_argument)
            else:
                separated_arguments.append(argument)

        return separated_arguments

    def construct_qubo(self, penalty: list[list], variables: list[str]) -> np.ndarray:
        """
        Creates QUBO matrix Q to solve linear problem of the form x^T * Q + x.

        :param penalty: List of lists containing all non-zero elements of the QUBO matrix as strings
        :param variables: Listing of all variables used in the problem
        :return: QUBO in numpy array format
        """
        # Create empty qubo matrix
        count_variables = len(variables)
        qubo = np.zeros((count_variables, count_variables))

        # Iterate through all the variables twice (x^T, x)
        for col, variable in enumerate(variables):
            for row, variable2 in enumerate(variables):
                # Save the parameters (values in the qubo)
                parameter = 0
                for argument in penalty:
                    if isinstance(argument, list):
                        # squared variables in diagonals (x^2 == x)
                        if (
                            len(argument) == 2
                            and any(isinstance(elem, str) and variable in elem for elem in argument)
                            and col == row
                        ):
                            parameter += argument[0]
                        # Multiplication of different variables not on diagonal
                        if (
                            len(argument) == 3
                            and variable in argument and variable2 in argument and variable > variable2
                        ):
                            parameter += argument[0]
                            # This value is already taking into account the factor 2 from quadratic term
                            # For the variables on the diagonal, if the parameter is zero
                            # We still have to check the sign in
                            # front of the decision variable. If it is "-", we have to put "-1" on the diagonal.
                    elif (isinstance(argument, str) and variable in argument
                          and variable2 in argument and variable == variable2):
                        if "-" in argument:
                            parameter += -1

                qubo[col, row] = parameter

        # Minimization problem
        qubo = -qubo.astype(int)

        return qubo

    def map(self, problem: dict, config: Config) -> tuple[dict, float]:
        """
        Converts linear program created with pulp to quadratic program to Ising with qiskit to QUBO matrix.

        :param problem: Dict containing the problem parameters
        :param config: Config with the parameters specified in Config class
        :return: Dict with the QUBO, time it took to map it
        """
        start = start_time_measurement()

        # Map Linear problem from dictionary (generated by pulp) to quadratic program to QUBO
        qp = self.map_pulp_to_qiskit(problem)
        logging.info(qp.export_as_lp_string())

        ineq2eq = InequalityToEquality()
        qp_eq = ineq2eq.convert(qp)

        int2bin = IntegerToBinary()
        qp_eq_bin = int2bin.convert(qp_eq)

        lineq2penalty = LinearEqualityToPenalty(100)
        qubo = lineq2penalty.convert(qp_eq_bin)

        variables = [variable.name for variable in qubo.variables]

        # convert penalty term to string to QUBO
        qubo_string = str(qubo.objective)
        arguments = self.convert_string_to_arguments(qubo_string)
        qubo_matrix = self.construct_qubo(arguments, variables)

        self.global_variables = variables

        return {"Q": qubo_matrix}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the ACL class for validation/evaluation.

        :param solution: bit_string containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()

        result = {"status": [0]}
        objective_value = 0
        variables = {}
        for bit in solution:
            if solution[bit] > 0 and "x" in self.global_variables[bit]:
                # We only care about assignments of vehicles to platforms:
                # We map the solution to the original variables
                variables[self.global_variables[bit]] = solution[bit]
                result["status"] = 'Optimal'  # TODO: I do not think every solution with at least one car is optimal
                objective_value += solution[bit]

        result["variables"] = variables
        result["obj_value"] = objective_value

        return result, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys
import argparse
import logging
import json
from collections.abc import Iterable
import yaml

from quark.Installer import Installer
from quark.utils import _expand_paths
from quark.utils_mpi import MPIStreamHandler, MPIFileHandler, get_comm

comm = get_comm()

# Add the paths
install_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(install_dir)

# The following line is currently needed for the hybrid jobs repo
sys.path.insert(1, os.path.dirname(os.path.realpath(__file__)))


def _filter_comments(file: Iterable) -> str:
    """
    Returns the content of the filehandle, ignoring all lines starting with '#'.

    :param file: file to be read
    :return: file content without comment lines
    """
    lines = []
    for line in file:
        if line.strip().startswith("#"):
            continue
        lines.append(line)
    return "".join(lines)


def setup_logging() -> None:
    """
    Sets up the logging.
    """
    logging.root.handlers = []
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
        handlers=[
            MPIFileHandler("logging.log"),
            MPIStreamHandler()
        ]
    )

    logging.info(" ============================================================ ")
    logging.info(r"             ___    _   _      _      ____    _  __           ")
    logging.info(r"            / _ \  | | | |    / \    |  _ \  | |/ /           ")
    logging.info(r"           | | | | | | | |   / _ \   | |_) | | ' /            ")
    logging.info(r"           | |_| | | |_| |  / ___ \  |  _ <  | . \            ")
    logging.info(r"            \__\_\  \___/  /_/   \_\ |_| \_\ |_|\_\           ")
    logging.info("                                                              ")
    logging.info(" ============================================================ ")
    logging.info("  A Framework for Quantum Computing Application Benchmarking  ")
    logging.info("                                                              ")
    logging.info("        Licensed under the Apache License, Version 2.0        ")
    logging.info(" ============================================================ ")


def start_benchmark_run(config_file: str = None, store_dir: str = None,
                        fail_fast: bool = False) -> None:
    """
    Starts a benchmark run from the code.
    """
    # TODO not reachable? dead code?
    setup_logging()

    # Helper for hybrid jobs
    if not config_file:
        config_file = os.environ["AMZN_BRAKET_HP_FILE"]
    if not store_dir:
        store_dir = os.environ["AMZN_BRAKET_JOB_RESULTS_DIR"]

    # Reads the config file
    with open(config_file, "r") as f:
        benchmark_config = json.load(f)

    benchmark_config = json.loads(benchmark_config["config"])

    from quark.BenchmarkManager import BenchmarkManager  # pylint: disable=C0415
    from quark.ConfigManager import ConfigManager  # pylint: disable=C0415

    config_manager = ConfigManager()
    config_manager.set_config(benchmark_config)

    benchmark_manager = BenchmarkManager(fail_fast=fail_fast)

    # Can be overridden by using the -m|--modules option
    installer = Installer()
    app_modules = installer.get_env(installer.get_active_env())
    benchmark_manager.orchestrate_benchmark(
        config_manager, store_dir=store_dir, app_modules=app_modules
    )


def create_benchmark_parser(parser: argparse.ArgumentParser):
    parser.add_argument("-c", "--config", help="Provide valid config file instead of interactive mode")
    parser.add_argument('-cc', '--createconfig', help='If you want o create a config without executing it',
                        required=False, action=argparse.BooleanOptionalAction)
    parser.add_argument('-s', '--summarize', nargs='+', help='If you want to summarize multiple experiments',
                        required=False)
    parser.add_argument('-m', '--modules', help="Provide a file listing the modules to be loaded")
    parser.add_argument('-rd', '--resume-dir', nargs='?', help='Provide results directory of the job to be resumed')
    parser.add_argument('-ff', '--failfast', help='Flag whether a single failed benchmark run causes QUARK to fail',
                        required=False, action=argparse.BooleanOptionalAction)

    parser.set_defaults(goal='benchmark')


def create_env_parser(parser: argparse.ArgumentParser):
    subparsers = parser.add_subparsers(help='env')

    parser_env = subparsers.add_parser('env', help='When you want to change something about the QUARK module env.')
    parser_env.add_argument('-c', '--configure', help='Configure certain QUARK modules', required=False)
    parser_env.add_argument('-a', '--activate', help='Activate a certain set of modules', required=False)
    parser_env.add_argument('-cm', '--createmoduledb', help='Create module db', required=False,
                            action=argparse.BooleanOptionalAction)
    parser_env.add_argument('-l', '--list', help='List existing environments', required=False,
                            action=argparse.BooleanOptionalAction)
    parser_env.add_argument('-s', '--show', help='Show the content of an env', required=False)
    parser_env.set_defaults(goal='env')


def handle_benchmark_run(args: argparse.Namespace) -> None:
    """
    Handles the different options of a benchmark run.

    :param args: Namespace with the arguments given by the user
    """
    from quark.BenchmarkManager import BenchmarkManager  # pylint: disable=C0415
    from quark.Plotter import Plotter  # pylint: disable=C0415

    benchmark_manager = BenchmarkManager(fail_fast=args.failfast)

    if args.summarize:
        benchmark_manager.summarize_results(args.summarize)
    else:
        from quark.ConfigManager import ConfigManager  # pylint: disable=C0415
        config_manager = ConfigManager()
        if args.modules:
            logging.info(f"Load application modules configuration from {args.modules}")
            # Preprocesses the 'modules' configuration:
            #   + Filters comment lines (lines starting with '#')
            #   + Replaces relative paths by taking them relative to the location of the modules configuration file
            base_dir = os.path.dirname(args.modules)
            with open(args.modules) as filehandler:
                app_modules = _expand_paths(json.loads(
                    _filter_comments(filehandler)), base_dir
                )
        else:
            # Gets current env here
            installer = Installer()
            app_modules = installer.get_env(installer.get_active_env())

        if args.config or args.resume_dir:
            if not args.config:
                args.config = os.path.join(args.resume_dir, "config.yml")
            logging.info(f"Provided config file at {args.config}")
            # Loads config
            with open(args.config) as filehandler:
                # Returns JSON object as a dictionary
                try:
                    benchmark_config = yaml.load(filehandler, Loader=yaml.FullLoader)
                except Exception as e:
                    logging.exception("Problem loading the given config file")
                    raise ValueError("Config file needs to be a valid QUARK YAML Config!") from e

                config_manager.set_config(benchmark_config)
        else:
            config_manager.generate_benchmark_configs(app_modules)

        if args.createconfig:
            logging.info("Selected config is:")
            config_manager.print()
        else:
            interrupted_results_path = None if args.resume_dir is None else os.path.join(
                args.resume_dir, "results.json"
            )
            benchmark_manager.orchestrate_benchmark(
                config_manager, app_modules,
                interrupted_results_path=interrupted_results_path
            )
            comm.Barrier()
            if comm.Get_rank() == 0:
                results = benchmark_manager.load_results()
                Plotter.visualize_results(results, benchmark_manager.store_dir)


def handler_env_run(args: argparse.Namespace) -> None:
    """
    Orchestrates the requests to the QUARK module environment.

    :param args: Namespace with the arguments given by the user
    """
    installer = Installer()
    if args.createmoduledb:
        installer.create_module_db()
    elif args.activate:
        installer.set_active_env(args.activate)
    elif args.configure:
        installer.configure(args.configure)
    elif args.list:
        installer.list_envs()
    elif args.show:
        installer.show(installer.get_env(args.show))


def start() -> None:
    """
    Main function that triggers the benchmarking process
    """
    setup_logging()

    try:
        parser = argparse.ArgumentParser()
        create_benchmark_parser(parser)
        create_env_parser(parser)

        args = parser.parse_args()
        if args.goal == "env":
            handler_env_run(args)

        else:
            handle_benchmark_run(args)

        logging.info(" ============================================================ ")
        logging.info(" ====================  QUARK finished!   ==================== ")
        logging.info(" ============================================================ ")

    except Exception as error:
        logging.error(error)
        raise error



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Union
import logging
from time import perf_counter
import numpy as np

from qiskit import QuantumCircuit, transpile
from qiskit.circuit import Parameter
from qiskit.providers import Backend
from qiskit_ibm_runtime.fake_provider import FakeProviderForBackendV2
from qiskit_aer import Aer, AerSimulator
from qiskit_aer.noise import NoiseModel

from quark.modules.applications.qml.generative_modeling.training.QCBM import QCBM
from quark.modules.applications.qml.generative_modeling.training.Inference import Inference
from quark.modules.applications.qml.generative_modeling.mappings.LibraryGenerative import LibraryGenerative

logging.getLogger("NoisyQiskit").setLevel(logging.WARNING)


class PresetQiskitNoisyBackend(LibraryGenerative):
    """
    This module maps a library-agnostic gate sequence to a qiskit circuit.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("PresetQiskitNoisyBackend")
        self.submodule_options = ["QCBM", "Inference"]

    circuit_transpiled = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "qiskit", "version": "1.3.0"},
            {"name": "qiskit_ibm_runtime", "version": "0.33.2"},
            {"name": "qiskit_aer", "version": "0.15.1"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for the Qiskit Library.

            :return: Dictionary with configurable settings.
            .. code-block:: python

                {
                "backend": {
                    "values": ["aer_simulator_gpu", "aer_simulator_cpu"],
                    "description": "Which backend do you want to use? "
                                   "In the NoisyQiskit Module only aer_simulators can be used."
                },

                "simulation_method": {
                    "values": ["automatic", "statevector", "density_matrix", "cpu_mps"],  # TODO Change names
                    "description": "What simulation methode should be used"
                },

                "n_shots": {
                    "values": [100, 1000, 10000, 1000000],
                    "description": "How many shots do you want use for estimating the PMF of the model?"
                },

                "transpile_optimization_level": {
                    "values": [1, 2, 3, 0],
                    "description": "Switch between different optimization levels in the Qiskit transpile routine. "
                                   "1: light optimization, 2: heavy optimization, 3: even heavier optimization, "
                                   "0: no optimization. Level 1 recommended as standard option."
                },

                "noise_configuration": {
                    "values": value_list,
                    "description": "What noise configuration do you want to use?"
                }
            }
        """
        provider = FakeProviderForBackendV2()
        backends = provider.backends()
        value_list = []
        value_list.append('No noise')
        for backend in backends:
            if backend.num_qubits >= 6:
                value_list.append(f'{backend.name} V{backend.version} {backend.num_qubits} Qubits')

        return {
            "backend": {
                "values": ["aer_simulator_gpu", "aer_simulator_cpu"],
                "description": "Which backend do you want to use? "
                               "In the NoisyQiskit Module only aer_simulators can be used."
            },

            "simulation_method": {
                "values": ["automatic", "statevector", "density_matrix", "cpu_mps"],  # TODO Change names
                "description": "What simulation methode should be used"
            },

            "n_shots": {
                "values": [100, 1000, 10000, 1000000],
                "description": "How many shots do you want use for estimating the PMF of the model?"
            },

            "transpile_optimization_level": {
                "values": [1, 2, 3, 0],
                "description": "Switch between different optimization levels in the Qiskit transpile routine. "
                               "1: light optimization, 2: heavy optimization, 3: even heavier optimization, "
                               "0: no optimization. Level 1 recommended as standard option."
            },

            "noise_configuration": {
                "values": value_list,
                "description": "What noise configuration do you want to use?"
            }
        }

    def get_default_submodule(self, option: str) -> Union[QCBM, Inference]:
        """
        Returns the default submodule based on the given option.

        :param option: The submodule option to select
        :return: Instance of the selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "QCBM":
            return QCBM()
        elif option == "Inference":
            return Inference()
        else:
            raise NotImplementedError(f"Option {option} not implemented")

    def sequence_to_circuit(self, input_data: dict) -> dict:
        """
        Maps the gate sequence, that specifies the architecture of a quantum circuit
        to its Qiskit implementation.

        :param input_data: Collected information of the benchmarking process
        :return: Same dictionary but the gate sequence is replaced by it Qiskit implementation
        """
        # TODO: Identical to CustomQiskitNoisyBackend.sequence_to_circuit -> move to Library
        n_qubits = input_data["n_qubits"]
        gate_sequence = input_data["gate_sequence"]
        circuit = QuantumCircuit(n_qubits, n_qubits)
        param_counter = 0
        for gate, wires in gate_sequence:
            if gate == "Hadamard":
                circuit.h(wires[0])
            elif gate == "X":
                circuit.x(wires[0])
            elif gate == "SX":
                circuit.sx(wires[0])
            elif gate == "RZ_PI/2":
                circuit.rz(np.pi / 2, wires[0])
            elif gate == "CNOT":
                circuit.cx(wires[0], wires[1])
            elif gate == "ECR":
                circuit.ecr(wires[0], wires[1])
            elif gate == "RZ":
                circuit.rz(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RX":
                circuit.rx(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RY":
                circuit.ry(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RXX":
                circuit.rxx(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "RYY":
                circuit.ryy(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "RZZ":
                circuit.rzz(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "CRY":
                circuit.cry(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "Barrier":
                circuit.barrier()
            elif gate == "Measure":
                circuit.measure(wires[0], wires[0])
            else:
                raise NotImplementedError(f"Gate {gate} not implemented")

        input_data["circuit"] = circuit
        input_data.pop("gate_sequence")
        logging.info(param_counter)
        input_data["n_params"] = len(circuit.parameters)
        return input_data

    @staticmethod
    def select_backend(config: str, n_qubits: int) -> Backend:
        """
        This method configures the backend.

        :param config: Name of a backend
        :param n_qubits: Number of qubits
        :return: Configured qiskit backend
        """
        # TODO: Identical to CustomQiskitNoisyBackend.select_backend -> move to Library
        if config == "aer_simulator_gpu":
            backend = Aer.get_backend("aer_simulator")
            backend.set_options(device="GPU")
        elif config == "aer_simulator_cpu":
            backend = Aer.get_backend("aer_simulator")
            backend.set_options(device="CPU")
        else:
            raise NotImplementedError(f"Device Configuration {config} not implemented")

        return backend

    def get_execute_circuit(self, circuit: QuantumCircuit, backend: Backend,  # pylint: disable=W0221
                            config: str, config_dict: dict) -> tuple[any, any]:
        """
        This method combines the qiskit circuit implementation and the selected backend and returns a function,
        that will be called during training.

        :param circuit: Qiskit implementation of the quantum circuit
        :param backend: Configured qiskit backend
        :param config: Name of a backend
        :param config_dict: Contains information about config
        :return: Tuple that contains a method that executes the quantum circuit for a given set of parameters and the
        transpiled circuit
        """
        # TODO: Identical to CustomQiskitNoisyBackend.get_execute_circuit -> move to Library
        n_shots = config_dict["n_shots"]
        n_qubits = circuit.num_qubits
        start = perf_counter()

        backend = self.decompile_noisy_config(config_dict, n_qubits)
        logging.info(f'Backend in Use: {backend=}')
        optimization_level = self.get_transpile_routine(config_dict['transpile_optimization_level'])
        seed_transp = 42  # Remove seed if wanted
        logging.info(f'Using {optimization_level=} with seed: {seed_transp}')
        circuit_transpiled = transpile(circuit, backend=backend, optimization_level=optimization_level,
                                       seed_transpiler=seed_transp)
        logging.info(f'Circuit operations before transpilation: {circuit.count_ops()}')
        logging.info(f'Circuit operations after transpilation: {circuit_transpiled.count_ops()}')
        logging.info(perf_counter() - start)

        if config in ["aer_simulator_cpu", "aer_simulator_gpu"]:
            def execute_circuit(solutions):
                all_circuits = [circuit_transpiled.assign_parameters(solution) for solution in solutions]
                jobs = backend.run(all_circuits, shots=n_shots)
                samples_dictionary = [jobs.result().get_counts(c).int_outcomes() for c in all_circuits]
                samples = []
                for result in samples_dictionary:
                    target_iter = np.zeros(2 ** n_qubits)
                    result_keys = list(result.keys())
                    result_vals = list(result.values())
                    target_iter[result_keys] = result_vals
                    target_iter = np.asarray(target_iter)
                    samples.append(target_iter)
                samples = np.asarray(samples)
                pmfs = samples / n_shots
                return pmfs, samples

        else:
            logging.error(f"Unknown backend option selected: {config}")
            logging.error("Run terminates with Exception error.")
            raise Exception

        return execute_circuit, circuit_transpiled

    @staticmethod
    def split_string(s):
        return s.split(' ', 1)[0]

    def decompile_noisy_config(self, config_dict: dict, num_qubits: int) -> Backend:
        """
        This method processes a configuration dictionary.
        If a custom noise configuration is specified, it creates a custom backend configuration; otherwise, it defaults
        to the 'aer_simulator' backend. It returns the configured backend.

        :param config_dict: Contains information about config
        :param num_qubits: Number of qubits
        :return: Configured qiskit backend
        """
        backend_config = config_dict['backend']
        device = 'GPU' if 'gpu' in backend_config else 'CPU'
        simulation_method, device = self.get_simulation_method_and_device(device, config_dict['simulation_method'])
        backend = self.select_backend_configuration(config_dict['noise_configuration'], num_qubits)

        self.configure_backend(backend, device, simulation_method)
        self.log_backend_info(backend)

        return backend

    def select_backend_configuration(self, noise_configuration: str, num_qubits: int) -> Backend:
        """
        This method selects the backend configuration based on the provided noise configuration.

        :param noise_configuration: Noise configuration type
        :param num_qubits: Number of qubits
        :return: Selected backend configuration
        """
        if "fake" in noise_configuration:
            return self.get_FakeBackend(noise_configuration, num_qubits)
        elif noise_configuration == "No noise":
            return Aer.get_backend("aer_simulator")
        elif noise_configuration in ['ibm_brisbane 127 Qubits', 'ibm_osaka 127 Qubits']:
            logging.warning("Not yet implemented. Please check upcoming QUARK versions.")
            raise ValueError(f"Noise configuration '{noise_configuration}' not yet implemented.")
        else:
            raise ValueError(f"Unknown noise configuration: {noise_configuration}")

    def configure_backend(self, backend: Backend, device: str, simulation_method: str) -> None:
        """
        This method configures the backend with the specified device and simulation method.

        :param backend: Backend to be configured
        :param device: Device type (CPU/GPU)
        :param simulation_method: Simulation method
        """
        backend.set_options(device=device)
        backend.set_options(method=simulation_method)

    def log_backend_info(self, backend: Backend):
        logging.info(f'Backend configuration: {backend.configuration()}')
        logging.info(f'Simulation method: {backend.options.method}')

    def get_simulation_method_and_device(self, device: str, simulation_config: str) -> tuple[str, str]:
        """
        This method determines the simulation method and device based on the provided configuration.

        :param device: Contains information about processing unit
        :param simulation_config: Contains information about qiskit simulation method
        :return: Tuple containing the simulation method and device
        """
        simulation_methods = {
            "statevector": "statevector",
            "density_matrix": "density_matrix",
            "cpu_mps": "matrix_product_state"
        }
        simulation_method = simulation_methods.get(simulation_config, 'automatic')
        if simulation_config == "cpu_mps":
            device = 'CPU'
        return simulation_method, device

    def get_transpile_routine(self, transpile_config: int) -> int:
        """
        This method returns the transpile routine based on the provided configuration.

        :param transpile_config: Configuration for transpile routine
        :return: Transpile routine level
        """
        return transpile_config if transpile_config in [0, 1, 2, 3] else 1

    def get_FakeBackend(self, noise_configuration: str, num_qubits: int) -> Backend:
        """
        This method returns a fake backend based on the provided noise configuration and number of qubits.

        :param noise_configuration: Noise configuration type
        :param num_qubits: Number of qubits
        :return: Fake backend simulator
        """
        backend_name = str(self.split_string(noise_configuration))
        provider = FakeProviderForBackendV2()
        try:
            backend = provider.backend(name=backend_name)
        except TypeError:
            logging.info("qiskit.providers.fake_provider.FakeProviderForBackendV2.get_backend overwritten. "
                         "Will be addressed with upcoming qiskit upgrade.")
            filtered_backends = [backend for backend in provider._backends if  # pylint: disable=W0212
                                 backend.name == backend_name]
            if not filtered_backends:
                raise FileNotFoundError()  # pylint: disable=W0707
            backend = filtered_backends[0]

        if num_qubits > backend.num_qubits:
            logging.warning(f'Requested number of qubits ({num_qubits}) exceeds the backend capacity. '
                            f'Using default aer_simulator.')
            return Aer.get_backend("aer_simulator")

        noise_model = NoiseModel.from_backend(backend)
        logging.info(f'Using {backend_name} with coupling map: {backend.coupling_map}')
        logging.info(f'Using {backend_name} with noise model: {noise_model}')
        return AerSimulator.from_backend(backend)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ast
import inspect
import json
import logging
import types
from collections import Counter
from functools import partial, wraps
from time import time
from typing import TypedDict

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pennylane as qml
from pennylane import numpy as npqml

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement

matplotlib.use('Agg')


class PennylaneQAOA(Solver):
    """
    Pennylane QAOA solver.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = [
            "arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            "arn:aws:braket:::device/quantum-simulator/amazon/tn1",
            "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony",
            "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3",
            "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy",
            "braket.local.qubit",
            "default.qubit",
            "default.qubit.autograd",
            "qulacs.simulator",
            "lightning.gpu",
            "lightning.qubit"
        ]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "pennylane", "version": "0.39.0"},
            {"name": "pennylane-lightning", "version": "0.39.0"},
            {"name": "amazon-braket-pennylane-plugin", "version": "1.30.2"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """

        if option == "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony":
            from quark.modules.devices.braket.Ionq import Ionq  # pylint: disable=C0415
            return Ionq("ionq", "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony")
        elif option == "arn:aws:braket:::device/quantum-simulator/amazon/sv1":
            from quark.modules.devices.braket.SV1 import SV1  # pylint: disable=C0415
            return SV1("SV1", "arn:aws:braket:::device/quantum-simulator/amazon/sv1")
        elif option == "arn:aws:braket:::device/quantum-simulator/amazon/tn1":
            from quark.modules.devices.braket.TN1 import TN1  # pylint: disable=C0415
            return TN1("TN1", "arn:aws:braket:::device/quantum-simulator/amazon/tn1")
        elif option == "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3":
            from quark.modules.devices.braket.Rigetti import Rigetti  # pylint: disable=C0415
            return Rigetti("Rigetti", "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3")
        elif option == "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy":
            from quark.modules.devices.braket.OQC import OQC  # pylint: disable=C0415
            return OQC("OQC", "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy")
        elif option == "braket.local.qubit":
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass("braket.local.qubit")
        elif option == "default.qubit":
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass("default.qubit")
        elif option == "default.qubit.autograd":
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass("default.qubit.autograd")
        elif option == "qulacs.simulator":
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass("qulacs.simulator")
        elif option == "lightning.gpu":
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass("lightning.gpu")
        elif option == "lightning.qubit":
            from quark.modules.devices.HelperClass import HelperClass  # pylint: disable=C0415
            return HelperClass("lightning.qubit")
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this solver.

        :return: Dictionary of configuration settings
        .. code-block:: python

                    return {
                            "shots": {  # number measurements to make on circuit
                                "values": list(range(10, 500, 30)),
                                "description": "How many shots do you need?"
                            },
                            "iterations": {  # number measurements to make on circuit
                                "values": [1, 10, 20, 50, 75],
                                "description": "How many iterations do you need?"
                            },
                            "layers": {
                                "values": [2, 3, 4],
                                "description": "How many layers for QAOA do you want?"
                            },
                            "coeff_scale": {
                                "values": [0.01, 0.1, 1, 10],
                                "description": "How do you want to scale your coefficients?"
                            },
                            "stepsize": {
                                "values": [0.0001, 0.001, 0.01, 0.1, 1],
                                "description": "Which stepsize do you want?"
                            }
                        }

        """
        return {
            "shots": {  # number measurements to make on circuit
                "values": [None] + list(range(10, 500, 30)),
                "description": "How many shots do you need?"
            },
            "iterations": {  # number measurements to make on circuit
                "values": [1, 10, 20, 50, 75],
                "description": "How many iterations do you need?"
            },
            "layers": {
                "values": [1, 2, 3, 4],
                "description": "How many layers for QAOA do you want?"
            },
            "coeff_scale": {
                "values": [0.01, 0.1, 1, 10],
                "description": "How do you want to scale your coefficients?"
            },
            "stepsize": {
                "values": [0.0001, 0.001, 0.01, 0.1, 1],
                "description": "Which stepsize do you want?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            shots: int
            depth: int
            iterations: int
            layers: int
            coeff_scale: float
            stepsize: float

        """
        shots: int
        depth: int
        iterations: int
        layers: int
        coeff_scale: float
        stepsize: float

    @staticmethod
    def normalize_data(data: any, scale: float = 1.0) -> any:
        """
        Not used currently, as I just scale the coefficients in the qaoa_operators_from_ising.

        :param data: Data to normalize
        :param scale: Scaling factor
        :return: Normalized data
        """
        return scale * data / np.max(np.abs(data))

    @staticmethod
    def qaoa_operators_from_ising(j: any, t: any, scale: float = 1.0) -> tuple[any, any]:
        """
        Generates pennylane cost and mixer Hamiltonians from the Ising matrix J and vector t.

        :param j: J matrix
        :param t: t vector
        :param scale: Scaling factor
        :return: Cost Hamiltonian and mixer Hamiltonian
        """
        # Define the scaling factor
        scaling_factor = scale * max(np.max(np.abs(j.flatten())), np.max(np.abs(t)))

        # Scale the coefficients
        j /= scaling_factor
        t /= scaling_factor

        sigzsigz_arr = [
            qml.PauliZ(i) @ qml.PauliZ(k) for i in range(len(j)) for k in range(len(j))
        ]

        sigz_arr = [qml.PauliZ(i) for i in range(len(t))]
        j_real = np.real(j.flatten())
        t_real = np.real(t)
        h_cost = qml.simplify(qml.Hamiltonian([*t_real, *j_real.flatten()], [*sigz_arr, *sigzsigz_arr]))

        # Definition of the mixer hamiltonian
        h_mixer = -1 * qml.qaoa.mixers.x_mixer(range(len(j)))

        return h_cost, h_mixer

    # pylint: disable=R0915
    def run(self, mapped_problem: dict, device_wrapper: any, config: Config, **kwargs: dict) -> tuple[any, float, dict]:
        """
        Runs Pennylane QAOA on the Ising problem.

        :param mapped_problem: Dict containing problem parameters mapped to the Ising model
        :param device_wrapper: Device to run the problem on
        :param config: QAOA solver settings
        :param kwargs: Contains store_dir for the plot of the optimization
        :return: Solution, the time it took to compute it and optional additional information
        """
        j = mapped_problem['J']
        t = mapped_problem['t']
        wires = j.shape[0]
        cost_h, mixer_h = self.qaoa_operators_from_ising(j, t, scale=config['coeff_scale'])

        # set up the problem
        try:
            device_arn = device_wrapper.arn
        except Exception as _:
            device_arn = None

        def qaoa_layer(gamma, alpha):
            qml.qaoa.cost_layer(gamma, cost_h)
            qml.qaoa.mixer_layer(alpha, mixer_h)

        def circuit(params, **kwargs):
            # circuit initialization
            for i in range(wires):
                qml.Hadamard(wires=i)
            # QAOA layers
            qml.layer(qaoa_layer, config['layers'], params[0], params[1])

        # The adjoint differentiation method is preferred over the default 'best' method for the lightning devices
        diff_method = "adjoint" if device_wrapper.device in ["lightning.qubit", "lightning.gpu"] else "best"
        # overwrite default with config
        if "differentiation_method" in config:
            diff_method = config["differentiation_method"]

        # TODO Make this interaction better with aws braket
        if device_arn is None:
            if device_wrapper.device == 'qulacs.simulator':
                dev = qml.device(device_wrapper.device, wires=wires, shots=config['shots'], gpu=True)
            elif device_wrapper.device in ['lightning.qubit', 'lightning.gpu']:
                # No number shots as diff method will be adjoint backprop for these devices
                if diff_method in ["adjoint", "backprop"]:
                    dev = qml.device(device_wrapper.device, wires=wires, shots=None, batch_obs=True)
                else:
                    dev = qml.device(device_wrapper.device, wires=wires, batch_obs=True, shots=config['shots'])
            elif device_wrapper.device == 'default.qubit':
                # No number shots as diff method will be adjoint backprop for these devices
                if diff_method in ["adjoint", "backprop"]:
                    dev = qml.device(device_wrapper.device, shots=None, wires=wires)
                else:
                    dev = qml.device(device_wrapper.device, wires=wires, shots=config['shots'])
            else:
                dev = qml.device(device_wrapper.device, wires=wires, shots=config['shots'])
        else:
            dev = qml.device(
                "braket.aws.qubit",
                device_arn=device_arn,
                wires=wires,
                s3_destination_folder=device_wrapper.s3_destination_folder,
                aws_session=device_wrapper.aws_session,
                parallel=True,
                shots=config['shots'],
                max_parallel=10,
                poll_timeout_seconds=6000,
            )

        @qml.qnode(dev, diff_method=diff_method)
        def cost_function(params):
            circuit(params)
            return qml.expval(cost_h)

        # To measure the QPU execution times we measure the timings of execute and batch_execute
        # in the pennylane devices
        # This is rather experimental and this has to be verified!
        dev.init_array = types.MethodType(monkey_init_array, dev)
        dev.init_array()
        real_decorator = partial(_pseudo_decor, device=dev)
        # TODO At some point validate whether execute and batch_execute are the only and best way to measure quantum
        #  execution times
        # TODO Check the impact of this ast.walk
        # We do this ast walk to assess whether execute is called in batch_execute,
        # which is the case for some simulators. This would distort the timing
        try:
            called_functions = [c.func.attr for c in ast.walk(ast.parse(inspect.getsource(dev.batch_execute).lstrip()))
                                if isinstance(c, ast.Call) and isinstance(c.func, ast.Attribute)]
        except AttributeError as e:
            logging.warning(f"{e} detected when trying to call pennylane.batch_execute.")
            logging.warning("Instead, tapes will be executed in serial.")
            called_functions = ['execute']
        dev.execute = real_decorator(dev.execute)
        if "execute" not in called_functions:
            dev.batch_execute = real_decorator(dev.batch_execute)

        # Initialize variational parameters randomly
        rand_params = np.random.uniform(size=[2, config['layers']])
        params = npqml.array(rand_params, requires_grad=True)
        logging.info(f"Starting params: {params}")

        # Optimization Loop
        optimizer = qml.GradientDescentOptimizer(stepsize=config['stepsize'])
        logging.info(f"Device: {device_wrapper.device}, Optimizer {optimizer}, Differentiation: {diff_method}, "
                     f"Optimization start...")

        additional_solver_information = {}
        min_param, min_cost, cost_pt, params_list, x = None, None, [], [], []
        run_id = round(time())
        start = start_time_measurement()

        for iteration in range(config['iterations']):
            t0 = start_time_measurement()
            # Evaluates the cost, then does a gradient step to new params
            try:
                params, cost_before = optimizer.step_and_cost(cost_function, params)
            except (ValueError, np.core._exceptions._ArrayMemoryError) as e:  # pylint: disable=W0212
                logging.error(f"When trying to run {optimizer} on {device_wrapper.device}, the following error was "
                              f"encountered")
                logging.error(e)
                logging.error("Run a smaller problem size or select another device.")
                raise e

            # Convert cost_before to a float, so it's easier to handle
            cost_before = float(cost_before)
            if iteration == 0:
                logging.info(f"Initial cost: {cost_before}")
            else:
                logging.info(f"Cost at step {iteration}: {cost_before}")

            # Log the current loss as a metric
            logging.info(f"Time to complete iteration {iteration + 1}: {end_time_measurement(t0)}")
            cost_pt.append(cost_before)
            params_list.append(params)
            x.append(iteration)

            if min_cost is None or min_cost > cost_before:
                min_cost = cost_before
                min_param = params

            if "store_dir" in kwargs:
                plt.figure(figsize=(6, 4))
                plt.plot(x, cost_pt, label='global minimum')
                plt.xlabel("Optimization steps")
                plt.ylabel("Cost / Energy")
                plt.title('/'.join([f"{key}: {value}" for (key, value) in config.items()]))
                plt.legend()
                plt.savefig(f"{kwargs['store_dir']}/plot_pennylane_qaoa_cost_{run_id}.pdf",
                            dpi=300)
                plt.clf()

        params = min_param
        logging.info(f"Final params: {params}")
        logging.info(f"Final costs: {min_cost}")

        @qml.qnode(dev)
        def samples(params):
            circuit(params)
            return [qml.sample(qml.PauliZ(i)) for i in range(wires)]

        def evaluate_params_sampling(params):
            s = samples([params[0], params[1]]).T
            s = (1 - s) / 2
            s = map(tuple, s)
            counts = Counter(s)
            indx = np.ndindex(*[2] * wires)
            probs = {p: counts.get(p, 0) / config['shots'] for p in indx}
            best_bitstring = max(probs, key=probs.get)

            return best_bitstring, probs

        @qml.qnode(dev)
        def probability_circuit(params):
            circuit(params)
            return qml.probs(wires=range(wires))

        def evaluate_params_probs(params):
            probs_raw = np.array(probability_circuit([params[0], params[1]]))
            indx = np.ndindex(*[2] * wires)
            probs = {p: probs_raw[i] for i, p in enumerate(indx)}
            best_bitstring = max(probs, key=probs.get)
            return best_bitstring, probs

        if config['shots'] is None or diff_method in ["backprop", "adjoint"]:
            best_bitstring, probs = evaluate_params_probs(params)
        else:
            best_bitstring, probs = evaluate_params_sampling(params)
        additional_solver_information["quantum_timings"] = dev.timings if len(dev.timings) > 0 else [
            -1]  # hack to make JSON parsing in R work
        additional_solver_information["quantum_timings_sum"] = sum(additional_solver_information["quantum_timings"])
        logging.info(f"{best_bitstring} with {probs[best_bitstring]}")

        logging.info(sorted(probs, key=probs.get, reverse=True)[:5])

        # Save the bitstring with the highest probability per iteration to bitstring_list
        bitstring_list = []
        for el in params_list:
            if config['shots'] is None or diff_method in ["backprop", "adjoint"]:
                bitstring, _ = evaluate_params_probs(el)
            else:
                bitstring, _ = evaluate_params_sampling(el)
            # bitstring, _ = evaluate_params_probs(el) if config['shots'] is None else evaluate_params_sampling(el)
            bitstring_list.append(bitstring)

        # Save the cost, best bitstring, variational parameters per iteration as well as the final prob. distribution
        # TODO: Maybe this can be done more efficient, e.g., only saving the circuit and its weights?
        json_data = {
            'cost': cost_pt,
            'bitstrings': bitstring_list,
            'params': [el.tolist() for el in params_list],  # convert list of tensors to list of lists
            'probs': {str(key): value for key, value in probs.items()}  # convert key (tuples to strings)
        }
        if "store_dir" in kwargs:
            with open(f"{kwargs['store_dir']}/qaoa_details_{run_id}.json", 'w') as fp:
                json.dump(json_data, fp)
        additional_solver_information["run_id"] = run_id
        return best_bitstring, end_time_measurement(start), additional_solver_information


def monkey_init_array(self):
    """
    Here we create the timings array where we later append the quantum timings.
    """
    self.timings = []


def _pseudo_decor(fun, device):
    """
    We use this decorator for measuring execute and batch_execute.
    """

    # Lift the name and doc of the function
    @wraps(fun)
    def ret_fun(*args, **kwargs):
        # Pre function execution here
        from time import time  # pylint: disable=W0621 disable=C0415 disable=W0404
        start_timing = time() * 1000
        returned_value = fun(*args, **kwargs)
        # Post execution here
        device.timings.append(round(time() * 1000 - start_timing, 3))
        return returned_value

    return ret_fun



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging
import numpy as np
from cma import CMAEvolutionStrategy
from tensorboardX import SummaryWriter
import matplotlib
from matplotlib import pyplot as plt
from matplotlib import figure, axes

from quark.modules.applications.qml.generative_modeling.training.TrainingGenerative import TrainingGenerative, Core, GPU
from quark.utils_mpi import is_running_mpi, get_comm

matplotlib.use('Agg')
MPI = is_running_mpi()
comm = get_comm()


class QCBM(TrainingGenerative):
    """
    This module optimizes the parameters of quantum circuit using CMA-ES.
    This training method is referred to as quantum circuit born machine (QCBM).
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("QCBM")

        self.n_states_range: list
        self.target: np.ndarray
        self.study_generalization: bool
        self.generalization_metrics: dict
        self.writer: SummaryWriter
        self.loss_func: callable
        self.fig: figure
        self.ax: axes.Axes

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "cma", "version": "4.0.0"},
            {"name": "matplotlib", "version": "3.9.3"},
            {"name": "tensorboard", "version": "2.18.0"},
            {"name": "tensorboardX", "version": "2.6.2.2"}
        ]

    def get_parameter_options(self) -> dict:
        """
        This function returns the configurable settings for the quantum circuit born machine.

        :return: Configuration settings for QCBM
        .. code-block:: python

            return {

                "population_size": {
                    "values": [5, 10, 100, 200, 10000],
                    "description": "What population size do you want?"
                },

                "max_evaluations": {
                    "values": [100, 1000, 20000, 100000],
                    "description": "What should be the maximum number of evaluations?"
                },

                "sigma": {
                    "values": [0.01, 0.5, 1, 2],
                    "description": "Which sigma would you like to use?"
                },

                "pretrained": {
                    "values": [False],
                    "custom_input": True,
                    "postproc": str,
                    "description": "Please provide the parameters of a pretrained model."
                },

                "loss": {
                    "values": ["KL", "NLL"],
                    "description": "Which loss function do you want to use?"
                }
            }
        """
        return {
            "population_size": {
                "values": [5, 10, 100, 200, 10000],
                "description": "What population size do you want?"
            },

            "max_evaluations": {
                "values": [100, 1000, 20000, 100000],
                "description": "What should be the maximum number of evaluations?"
            },

            "sigma": {
                "values": [0.01, 0.5, 1, 2],
                "description": "Which sigma would you like to use?"
            },

            "pretrained": {
                "values": [False],
                "custom_input": True,
                "postproc": str,
                "description": "Please provide the parameters of a pretrained model."
            },

            "loss": {
                "values": ["KL", "NLL"],
                "description": "Which loss function do you want to use?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            population_size: int
            max_evaluations: int
            sigma: float
            pretrained: str
            loss: str

        """
        population_size: int
        max_evaluations: int
        sigma: float
        pretrained: str
        loss: str

    def get_default_submodule(self, option: str) -> Core:
        """
        Raises ValueError as this module has no submodules.

        :param option: Option name
        :raises ValueError: If called, since this module has no submodules
        """
        raise ValueError("This module has no submodules.")

    def setup_training(self, input_data: dict, config: Config) -> tuple[float, dict]:
        """
        Method to configure the training setup including CMA-ES and tensorboard.

        :param input_data: A representation of the quantum machine learning model that will be trained
        :param config: Config specifying the parameters of the training
        :return: Random initial parameter and options for CMA-ES
        """

        logging.info(
            f"Running config: [backend={input_data['backend']}] "
            f"[n_qubits={input_data['n_qubits']}] "
            f"[population_size={config['population_size']}]"
        )

        self.study_generalization = "generalization_metrics" in list(input_data.keys())
        if self.study_generalization:
            self.generalization_metrics = input_data["generalization_metrics"]
            self.generalization_metrics.n_shots = input_data["n_shots"]

        self.writer = SummaryWriter(input_data["store_dir_iter"])

        if config['loss'] == "KL":
            self.loss_func = self.kl_divergence
        elif config['loss'] == "NLL":
            self.loss_func = self.nll
        elif config['loss'] == "MMD":
            self.loss_func = self.mmd
        else:
            raise NotImplementedError("Loss function not implemented")

        n_params = input_data["n_params"]
        x0 = (np.random.rand(n_params) - 0.5) * np.pi
        if config["pretrained"] != "False":
            parameters = np.load(config["pretrained"])
            x0[:len(parameters)] = parameters
            logging.info(f'Training starting from parameters in path {config["pretrained"]}')

        options = {
            'bounds': [n_params * [-np.pi], n_params * [np.pi]],
            'maxfevals': config['max_evaluations'],
            'popsize': config['population_size'],
            'verbose': -3,
            'tolfun': 1e-12
        }

        return x0, options

    def start_training(self, input_data: dict, config: dict, **kwargs: dict) -> dict:
        """
        This function finds the best parameters of the circuit on a transformed problem instance and returns a solution.

        :param input_data: A representation of the quantum machine learning model that will be trained
        :param config: Config specifying the parameters of the training
        :param kwargs: Optional additional settings
        :return: Dictionary including the information of previous modules as well as of the training
        """

        size = None
        input_data['MPI_size'] = size
        input_data["store_dir_iter"] += f"_{input_data['dataset_name']}_qubits{input_data['n_qubits']}"
        x0, options = self.setup_training(input_data, config)

        is_master = comm.Get_rank() == 0
        if is_master:
            self.target = np.asarray(input_data["histogram_train"])
            self.target[self.target == 0] = 1e-8

        self.n_states_range = range(2 ** input_data['n_qubits'])
        execute_circuit = input_data["execute_circuit"]
        timing = self.Timing()

        es = CMAEvolutionStrategy(x0.get() if GPU else x0, config['sigma'], options)

        for parameter in ["best_parameters", "time_circuit", "time_loss", "KL", "best_sample"]:
            input_data[parameter] = []

        best_loss = float("inf")
        self.fig, self.ax = plt.subplots()

        while not es.stop():
            solutions = es.ask()
            epoch = es.result[4]
            sigma = es.sigma

            timing.start_recording()
            pmfs_model, samples = execute_circuit(solutions)
            pmfs_model = np.asarray(pmfs_model)
            time_circ = timing.stop_recording()

            timing.start_recording()
            if is_master:
                loss_epoch = self.loss_func(pmfs_model.reshape([config['population_size'], -1]), self.target)
            else:
                loss_epoch = np.empty(config["population_size"])

            comm.Bcast(loss_epoch, root=0)
            comm.Barrier()

            time_loss = timing.stop_recording()

            es.tell(solutions, loss_epoch.get() if GPU else loss_epoch)

            if es.result[1] < best_loss:
                best_loss = es.result[1]
                best_pmf = self.data_visualization(loss_epoch, pmfs_model, samples, epoch)

            input_data["best_parameters"].append(es.result[0])
            input_data["KL"].append(float(es.result[1]))

            logging.info(
                f"[Iteration {epoch}] "
                f"[{config['loss']}: {es.result[1]:.5f}] "
                f"[Circuit processing: {(time_circ):.3f} ms] "
                f"[{config['loss']} processing: {(time_loss):.3f} ms] "
                f"[sigma: {sigma:.5f}]"
            )

        plt.close()
        self.writer.flush()
        self.writer.close()

        input_data["best_parameter"] = es.result[0]
        best_sample = self.sample_from_pmf(best_pmf.get() if GPU else best_pmf,  # pylint: disable=E0606
                                           n_shots=input_data["n_shots"])
        input_data["best_sample"] = best_sample.get() if GPU else best_sample  # pylint: disable=E1101

        return input_data

    def data_visualization(self, loss_epoch: np.ndarray, pmfs_model: np.ndarray, samples: any, epoch: int) -> (
            np.ndarray):
        """
        Visualizes the data and metrics for training.

        :param loss_epoch: Loss for the current epoch
        :param pmfs_model: The probability mass functions from the model
        :param samples: The samples from the model
        :param epoch: The current epoch number
        :return: Best probability mass function for visualization
        """
        index = loss_epoch.argmin()
        best_pmf = pmfs_model[index] / pmfs_model[index].sum()

        if self.study_generalization:
            if samples is None:
                counts = self.sample_from_pmf(
                    pmf=best_pmf.get() if GPU else best_pmf,
                    n_shots=self.generalization_metrics.n_shots)
            else:
                counts = samples[int(index)]

            metrics = self.generalization_metrics.get_metrics(counts if GPU else counts)
            for key, value in metrics.items():
                self.writer.add_scalar(f"metrics/{key}", value, epoch)

        nll = self.nll(best_pmf.reshape([1, -1]), self.target)
        kl = self.kl_divergence(best_pmf.reshape([1, -1]), self.target)
        mmd = self.mmd(best_pmf.reshape([1, -1]), self.target)

        self.writer.add_scalar("metrics/NLL", nll.get() if GPU else nll, epoch)
        self.writer.add_scalar("metrics/KL", kl.get() if GPU else kl, epoch)
        self.writer.add_scalar("metrics/MMD", mmd.get() if GPU else mmd, epoch)

        self.ax.clear()
        self.ax.imshow(
            best_pmf.reshape(int(np.sqrt(best_pmf.size)), int(np.sqrt(best_pmf.size))).get()
            if GPU else best_pmf.reshape(int(np.sqrt(best_pmf.size)), int(np.sqrt(best_pmf.size))),
            cmap='binary',
            interpolation='none'
        )
        self.ax.set_title(f'Iteration {epoch}')
        self.writer.add_figure('grid_figure', self.fig, global_step=epoch)

        return best_pmf



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict

import networkx as nx
import pulser

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class NeutralAtom(Mapping):
    """
    Neutral atom formulation for MIS.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["NeutralAtomMIS"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of requirements of this module
        """
        return [{"name": "pulser", "version": "1.1.1"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Empty dictionary, as this mapping has no configurable settings
        """
        return {}

    class Config(TypedDict):
        """
        Configuration options for Neutral Atom MIS mapping.
        """
        pass

    def map(self, problem: nx.Graph, config: Config) -> tuple[dict, float]:
        """
        Maps the networkx graph to a neutral atom MIS problem.

        :param problem: Networkx graph representing the MIS problem
        :param config: Config with the parameters specified in Config class
        :return: Tuple containing a dictionary with the neutral MIS and time it took to map it
        """
        start = start_time_measurement()

        pos = nx.get_node_attributes(problem, 'pos')
        register = pulser.Register(pos)

        neutral_atom_problem = {
            'graph': problem,
            'register': register
        }

        return neutral_atom_problem, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "NeutralAtomMIS":
            from quark.modules.solvers.NeutralAtomMIS import NeutralAtomMIS  # pylint: disable=C0415
            return NeutralAtomMIS()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



import unittest
from unittest.mock import patch, MagicMock

from quark.modules.Core import Core
from quark.ConfigManager import ConfigManager


class TestConfigManager(unittest.TestCase):

    def setUp(self):
        self.config_manager = ConfigManager()

    @patch("src.ConfigManager.inquirer.prompt")
    @patch("src.ConfigManager.checkbox")
    def test_query_module(self, mock_checkbox, mock_prompt):
        # Mock responses for checkbox and prompt
        mock_checkbox.return_value = {"param1": [1, 2]}  # Simulates a user selecting 1 and 2
        mock_prompt.return_value = {"param2": "a"}  # Simulates a user selecting 'a' for param2

        param_opts = {
            "param1": {"values": [1, 2, 3], "description": "Test parameter 1"},
            "param2": {"values": ["a", "b"], "description": "Test parameter 2", "exclusive": True},
        }

        config = ConfigManager._query_for_config(param_opts)

        # Assert the results
        self.assertEqual(config["param1"], [1, 2])

    def test_set_config(self):
        config = {
            "application": {
                "name": "TestApp",
                "config": {"param1": 1},
                "submodules": []
            },
            "repetitions": 2
        }
        self.config_manager.set_config(config)

        self.assertEqual(self.config_manager.config["application"]["name"], "TestApp")
        self.assertEqual(self.config_manager.config["repetitions"], 2)

    def test_is_legacy_config(self):
        legacy_config = {"mapping": {"TestMapping": {}}}
        modern_config = {"application": {"name": "TestApp"}, "repetitions": 2}

        self.assertTrue(ConfigManager.is_legacy_config(legacy_config))
        self.assertFalse(ConfigManager.is_legacy_config(modern_config))

    def test_translate_legacy_config_missing_device(self):
        # Mock a legacy config
        config = {
            "application": {"name": "TestApp", "config": {}},
            "repetitions": 1,
            "mapping": {
                "direct": {"solver": [{"name": "DirectSolver", "config": {}, "device": []}]},
                "solver1": {
                    "config": {"param1": "value1"},
                    "solver": [{"name": "Solver1", "config": {"param2": "value2"}, "device": []}],
                },
            },
        }

        # Call the translate_legacy_config method
        translated_config = ConfigManager.translate_legacy_config(config)

        # Assertions to verify the output
        self.assertIn("application", translated_config)
        self.assertIn("repetitions", translated_config)
        self.assertEqual(translated_config["repetitions"], 1)
        self.assertIn("submodules", translated_config["application"])
        self.assertEqual(len(translated_config["application"]["submodules"]), 2)

        # Verify the structure of the submodules
        direct_module = translated_config["application"]["submodules"][0]
        self.assertEqual(len(direct_module["submodules"]), 0)

        solver1_module = translated_config["application"]["submodules"][1]
        self.assertEqual(solver1_module["name"], "solver1")
        self.assertEqual(solver1_module["config"]["param1"], "value1")
        self.assertEqual(len(solver1_module["submodules"]), 1)
        self.assertEqual(solver1_module["submodules"][0]["name"], "Solver1")
        self.assertEqual(solver1_module["submodules"][0]["config"]["param2"], "value2")

    def test_translate_legacy_config_helper(self):
        legacy_solver = {
            "solver": [
                {"name": "SolverA", "config": {}, "device": [{"name": "DeviceA", "config": {}}]}
            ]
        }
        result = ConfigManager.translate_legacy_config_helper(legacy_solver, "solver")

        # Assertions
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["name"], "SolverA")
        self.assertEqual(len(result[0]["submodules"]), 1)

    @patch("src.ConfigManager._get_instance_with_sub_options")
    def test_load_config(self, mock_get_instance):
        mock_app_instance = MagicMock()
        mock_get_instance.return_value = mock_app_instance

        config = {
            "application": {
                "name": "TestApp",
                "config": {"param1": 1},
                "submodules": []
            },
            "repetitions": 2
        }
        self.config_manager.set_config(config)

        # Call load_config
        self.config_manager.load_config([{"name": "TestApp"}])

        # Assertions
        self.assertEqual(self.config_manager.application, mock_app_instance)
        mock_get_instance.assert_called_once_with([{"name": "TestApp"}], "TestApp")

    def test_initialize_module_classes(self):
        parent_module = MagicMock(spec=Core)
        parent_module.get_submodule = MagicMock(side_effect=lambda name: MagicMock(name=name))

        config = {
            "name": "TestModule",
            "config": {"param1": "value1"},
            "submodules": [
                {"name": "SubModule1", "config": {}, "submodules": []},
                {"name": "SubModule2", "config": {}, "submodules": []},
            ],
        }

        result = ConfigManager.initialize_module_classes(parent_module, config)
        self.assertEqual(result["name"], "TestModule")
        self.assertEqual(len(result["submodules"]), 2)
        self.assertIn("instance", result)
        parent_module.get_submodule.assert_called_with("TestModule")

    def test_get_config(self):
        mock_config = {"application": {"name": "TestApp"}, "repetitions": 3}
        self.config_manager.config = mock_config
        self.assertEqual(self.config_manager.get_config(), mock_config)

    def test_get_app(self):
        mock_application = MagicMock()
        mock_config = {"application": {"instance": mock_application}}
        self.config_manager.config = mock_config
        self.assertEqual(self.config_manager.get_app(), mock_application)

    def test_get_reps(self):
        mock_config = {"repetitions": 5}
        self.config_manager.config = mock_config
        self.assertEqual(self.config_manager.get_reps(), 5)

    def test_start_create_benchmark_backlog(self):
        mock_application = {
            "name": "TestApp",
            "config": {"param1": [1, 2]},
            "submodules": [
                    {
                        "name": "SubModule1",
                        "config": {"sub_param": [3, 4]},
                        "submodules": [],
                        "instance": MagicMock(),
                    }
            ],
            "instance": MagicMock(),
        }
        self.config_manager.config = {"application": mock_application}
        backlog = self.config_manager.start_create_benchmark_backlog()
        self.assertEqual(len(backlog), 4)
        self.assertEqual(backlog[0]["name"], "TestApp")

    def test_create_benchmark_backlog(self):
        module = {
            "name": "TestModule",
            "config": {"param1": [1, 2]},
            "submodules": [
                    {
                        "name": "SubModule1",
                        "config": {"sub_param": [3, 4]},
                        "submodules": [],
                        "instance": MagicMock(),
                    }
            ],
            "instance": MagicMock(),
        }
        backlog = ConfigManager.create_benchmark_backlog(module)
        self.assertEqual(len(backlog), 4)
        self.assertEqual(backlog[0]["config"]["param1"], 1)
        self.assertEqual(backlog[1]["config"]["param1"], 1)

    def test_save(self):
        mock_config = {"application": {"name": "TestApp"}, "repetitions": 3}
        self.config_manager.config = mock_config
        with patch("builtins.open", unittest.mock.mock_open()) as mocked_open:
            self.config_manager.save("/mock/store/dir")
            mocked_open.assert_called_once_with("/mock/store/dir/config.yml", "w")

    def test_print(self):
        mock_config = {"application": {"name": "TestApp"}, "repetitions": 3}
        self.config_manager.config = mock_config
        with patch("yaml.dump") as mock_yaml_dump:
            self.config_manager.print()
            mock_yaml_dump.assert_called_once_with(mock_config)

    def test_create_tree_figure(self):
        mock_config = {
            "name": "TestApp",
            "submodules": [
                {"name": "SubModule1", "submodules": []},
                {"name": "SubModule2", "submodules": []},
            ],
        }
        self.config_manager.config = {"application": mock_config}
        with patch("matplotlib.pyplot.savefig") as mock_savefig:
            self.config_manager.create_tree_figure("/mock/store/dir")
            mock_savefig.assert_called_once_with("/mock/store/dir/BenchmarkGraph.png", format="PNG")

    @patch("src.ConfigManager.inquirer.prompt")
    @patch("src.ConfigManager.checkbox")
    def test_query_for_config(self, mock_checkbox, mock_prompt):
        mock_checkbox.return_value = {"param1": [1, 2, "Custom Input"]}
        mock_prompt.side_effect = [{"custom_input": "custom_value"}]
        param_opts = {
            "param1": {"values": [1, 2, 3], "description": "Test parameter 1", "custom_input": True}
        }
        config = ConfigManager._query_for_config(param_opts)
        self.assertIn("param1", config)
        self.assertEqual(config["param1"], [1, 2, "custom_value"])



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from quark.modules.Core import Core


class Mapping(Core, ABC):
    """
    This module translates the input data and problem specification from the parent module,
    e.g., the application into a mathematical formulation suitable the submodule, e.g., a solver.
    """

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Maps the data to the correct target format.

        :param input_data: Data which should be mapped
        :param config: Config of the mapping
        :param kwargs: Optional keyword arguments
        :return: Tuple with mapped problem and the time it took to map it
        """
        output, preprocessing_time = self.map(input_data, config)
        return output, preprocessing_time

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Reverse transformation/mapping from the submodule's format to the mathematical formulation
        suitable for the parent module.

        :param input_data: Data which should be reverse-mapped
        :param config: Config of the reverse mapping
        :param kwargs: Optional keyword arguments
        :return: Tuple with reverse-mapped problem and the time it took to map it
        """
        output, postprocessing_time = self.reverse_map(input_data)
        return output, postprocessing_time

    @abstractmethod
    def map(self, problem: any, config: dict) -> tuple[any, float]:
        """
        Maps the given problem into a specific format suitable for the submodule, e.g., a solver.

        :param config: Instance of class Config specifying the mapping settings
        :param problem: Problem instance which should be mapped to the target representation
        :return: Mapped problem and the time it took to map it
        """
        pass

    def reverse_map(self, solution: any) -> tuple[any, float]:
        """
        Maps the solution back to the original problem. This might not be necessary in all cases, so the default is
        to return the original solution. This might be needed to convert the solution to a representation needed
        for validation and evaluation.

        :param solution: Solution provided by submodule, e.g., the Solver class
        :return: Reverse-mapped solution and the time it took to create it
        """
        return solution, 0



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from abc import ABC
import time

try:
    import cupy as np
    GPU = True
    logging.info("Using CuPy, data processing on GPU")
except ModuleNotFoundError:
    import numpy as np
    GPU = False
    logging.info("CuPy not available, using vanilla numpy, data processing on CPU")

from quark.modules.Core import Core
from quark.modules.applications.qml.Training import Training
from quark.utils import start_time_measurement, end_time_measurement


class TrainingGenerative(Core, Training, ABC):
    """
    The Training module is the base class fot both finding (QCBM) and executing trained models (Inference).
    """

    def __init__(self, name: str):
        """
        Constructor method.

        :param name: Name of the training instance
        """
        self.name = name
        super().__init__()
        self.n_states_range = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: list of dict with requirements of this module
        """
        return [{"name": "numpy", "version": "1.26.4"}]

    def postprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[dict, float]:
        """
        Perform the actual training of the machine learning model.

        :param input_data: Collected information of the benchmarking process
        :param config: Training settings
        :param kwargs: Optional additional arguments
        :return: Training results and the postprocessing time
        """
        start = start_time_measurement()
        logging.info("Start training")
        training_results = self.start_training(
            self.preprocessed_input,
            config,
            **kwargs
        )
        for dict_key in ["backend", "circuit", "execute_circuit"]:
            training_results.pop(dict_key)
        postprocessing_time = end_time_measurement(start)
        logging.info(f"Training finished in {postprocessing_time / 1000} s.")
        return training_results, postprocessing_time

    def sample_from_pmf(self, pmf: np.ndarray, n_shots: int) -> np.ndarray:
        """
        This function samples from the probability mass function generated by the quantum circuit.

        :param pmf: Probability mass function generated by the quantum circuit
        :param n_shots: Number of shots
        :return: Number of counts in the 2**n_qubits bins
        """
        samples = np.random.choice(self.n_states_range, size=n_shots, p=pmf)
        counts = np.bincount(samples, minlength=len(self.n_states_range))
        return counts

    def kl_divergence(self, pmf_model: np.ndarray, pmf_target: np.ndarray) -> np.ndarray:
        """
        This function calculates the Kullback-Leibler divergence, that is used as a loss function.

        :param pmf_model: Probability mass function generated by the quantum circuit
        :param pmf_target: Probability mass function of the target distribution
        :return: Kullback-Leibler divergence
        """
        pmf_model[pmf_model == 0] = 1e-8
        return np.sum(pmf_target * np.log(pmf_target / pmf_model), axis=1)

    def nll(self, pmf_model: np.ndarray, pmf_target: np.ndarray) -> np.ndarray:
        """
        This function calculates th negative log likelihood, that is used as a loss function.

        :param pmf_model: Probability mass function generated by the quantum circuit
        :param pmf_target: Probability mass function of the target distribution
        :return: Negative log likelihood
        """
        pmf_model[pmf_model == 0] = 1e-8
        return -np.sum(pmf_target * np.log(pmf_model), axis=1)

    def mmd(self, pmf_model: np.ndarray, pmf_target: np.ndarray) -> np.ndarray:
        """
        This function calculates the maximum mean discrepancy, that is used as a loss function.

        :param pmf_model: Probability mass function generated by the quantum circuit
        :param pmf_target: Probability mass function of the target distribution
        :return: Maximum mean discrepancy
        """
        pmf_model[pmf_model == 0] = 1e-8
        sigma = 1 / pmf_model.shape[1]
        kernel_distance = np.exp((-np.square(pmf_model - pmf_target) / (sigma ** 2)))
        mmd = 2 - 2 * np.mean(kernel_distance, axis=1)
        return mmd

    class Timing:
        """
        This module is an abstraction of time measurement for both CPU and GPU processes.
        """

        def __init__(self):
            """
            Constructor method.
            """

            if GPU:
                self.start_cpu: time.perf_counter
            else:
                self.start_gpu: np.cuda.Event
                self.end_gpu: time.perf_counter

            self.start_recording = self.start_recording_gpu if GPU else self.start_recording_cpu
            self.stop_recording = self.stop_recording_gpu if GPU else self.stop_recording_cpu

        def start_recording_cpu(self) -> None:
            """
            This is a function to start time measurement on the CPU.
            """
            self.start_cpu = start_time_measurement()

        def stop_recording_cpu(self) -> float:
            """
            This is a function to stop time measurement on the CPU.

            .return: Elapsed time in milliseconds
            """
            return end_time_measurement(self.start_cpu)

        def start_recording_gpu(self) -> None:
            """
            This is a function to start time measurement on the GPU.
            """
            self.start_gpu = np.cuda.Event()
            self.end_gpu = np.cuda.Event()
            self.start_gpu.record()

        def stop_recording_gpu(self) -> float:
            """
            This is a function to stop time measurement on the GPU.

            :return: Elapsed time in milliseconds
            """
            self.end_gpu.record()
            self.end_gpu.synchronize()
            return np.cuda.get_elapsed_time(self.start_gpu, self.end_gpu)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
import logging
from pprint import pformat
from typing import TypedDict

import numpy as np

from quark.modules.applications.qml.generative_modeling.circuits.CircuitCardinality import CircuitCardinality
from quark.modules.applications.qml.generative_modeling.data.data_handler.DataHandlerGenerative import DataHandlerGenerative
from quark.modules.applications.qml.generative_modeling.metrics.MetricsGeneralization import MetricsGeneralization
from quark.utils import start_time_measurement, end_time_measurement


class DiscreteData(DataHandlerGenerative):
    """
    A data handler for discrete datasets with cardinality constraints.
    This class creates a dataset with a cardinality constraint and provides
    methods for generalization metrics computing and evaluation.
    """

    def __init__(self):
        super().__init__("")
        self.submodule_options = ["CircuitCardinality"]
        self.n_registers = None
        self.n_qubits = None
        self.train_size = None
        self.histogram_train = None
        self.histogram_solution = None
        self.generalization_metrics = None
        self.samples = None
        self.train_set = None
        self.solution_set = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "numpy", "version": "1.26.4"}]

    def get_default_submodule(self, option: str) -> CircuitCardinality:
        """
        Get the default submodule based on the given option.

        :param option: Submodule option
        :return: Corresponding submodule
        """
        if option == "CircuitCardinality":
            return CircuitCardinality()
        else:
            raise NotImplementedError(f"Circuit Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: A dictionary of parameter options
        .. code-block:: python

        return {
            "train_size": {
            "values": [0.1, 0.3, 0.5, 0.7, 0.9, 1.0],
            "description": "What percentage of the dataset do you want to use for training?"
            }
        }
        """
        return {
            "train_size": {
                "values": [0.1, 0.3, 0.5, 0.7, 0.9, 1.0],
                "description": "What percentage of the dataset do you want to use for training?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            train_size: int

        """

        train_size: int

    def data_load(self, gen_mod: dict, config: Config) -> dict:
        """
        The cardinality constrained dataset is created and split into a training set.

        :param gen_mod: Dictionary with collected information of the previous modules
        :param config: Config specifying the parameters of the data handler
        :return: Dictionary including the mapped problem
        """
        dataset_name = "Cardinality_Constraint"
        self.n_qubits = gen_mod["n_qubits"]
        self.train_size = config["train_size"]
        num_ones = self.n_qubits // 2

        # Generate all possible binary permutations of length n_qubits using NumPy arrays
        search_space = np.array(list(itertools.product([0, 1], repeat=self.n_qubits)))
        search_space = np.apply_along_axis(lambda x: ''.join(str(bit) for bit in x), 1, search_space)

        # Filter the binary permutations based on the cardinality constraint using np.char.count
        cardinality_constraint = np.char.count(search_space, '1') == num_ones
        solution_set = search_space[cardinality_constraint]

        # Use np.random.choice to select the desired number of rows randomly
        size_train = int(self.train_size * len(solution_set))
        train_set_indices = np.random.choice(len(solution_set), size_train, replace=False)
        train_set = solution_set[train_set_indices]

        # Create the histogram solution data
        self.histogram_solution = np.zeros(2 ** self.n_qubits)
        self.solution_set = np.array([int(i, 2) for i in solution_set])
        self.histogram_solution[self.solution_set] = 1 / len(self.solution_set)

        # Create the histogram training data
        self.histogram_train = np.zeros(2 ** self.n_qubits)
        self.train_set = np.array([int(i, 2) for i in train_set])
        self.histogram_train[self.train_set] = 1 / len(self.train_set)

        train_set_binary = np.array([list(map(int, s)) for s in train_set])
        solution_set_binary = np.array([list(map(int, s)) for s in solution_set])

        application_config = {
            "dataset_name": dataset_name,
            "binary_train": train_set_binary,
            "binary_solution": solution_set_binary,
            "train_size": self.train_size,
            "n_qubits": self.n_qubits,
            "n_registers": 2,
            "histogram_solution": self.histogram_solution,
            "histogram_train": self.histogram_train,
            "store_dir_iter": gen_mod["store_dir_iter"]
        }

        if self.train_size != 1:
            self.generalization_metrics = MetricsGeneralization(
                train_set=self.train_set,
                train_size=self.train_size,
                solution_set=self.solution_set,
                n_qubits=self.n_qubits,
            )
            application_config["generalization_metrics"] = self.generalization_metrics

        return application_config

    def generalization(self) -> tuple[dict, float]:
        """
        Calculate generalization metrics for the generated.

        :return: Tuple containing a dictionary of generalization metrics and the execution time
        """
        start = start_time_measurement()
        results = self.generalization_metrics.get_metrics(self.samples)

        logging.info(pformat(results))

        return results, end_time_measurement(start)

    def evaluate(self, solution: dict) -> tuple[dict, float]:
        """
        Evaluates a given solution and calculates the histogram of generated samples and the minimum KL divergence
        value.

        :param solution: Dictionary containing the solution data, including generated samples and KL divergence values
        :return: Tuple containing a dictionary with the histogram of generated samples and the minimum KL divergence
                 value, and the time it took to evaluate the solution
        """
        start = start_time_measurement()
        self.samples = solution["best_sample"]
        n_shots = np.sum(self.samples)

        histogram_generated = np.asarray(self.samples) / n_shots
        histogram_generated[histogram_generated == 0] = 1e-8

        kl_list = solution["KL"]
        kl_best = min(kl_list)

        evaluate_dict = {"histogram_generated": histogram_generated, "KL_best": kl_best}

        return evaluate_dict, end_time_measurement(start)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import combinations
from typing import TypedDict
import logging

from nnf import And

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class DinneenQUBO(Mapping):
    """
    QUBO formulation for SAT as given by Dinneen -- see also the description in the QUARK paper (2202.03028).
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]
        self.nr_vars = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "nnf", "version": "0.4.1"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary with parameter options
        .. code-block:: python

            return {
                    "lagrange": {
                        "values": [0.1, 1, 2],
                        "description": "What Lagrange param. to multiply with the number of (hard) constr.?"
                    }
                }
        """
        return {
            "lagrange": {
                "values": [0.1, 1, 2],
                "description": "What Lagrange parameter to multiply with the number of (hard) constraints?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             lagrange: float

        """
        lagrange: float

    def map(self, problem: tuple[And, list], config: Config) -> tuple[dict, float]:
        """
        Performs the mapping into a QUBO formulation, as given by Dinneen. See also the QUARK paper.

        :param problem: SAT problem
        :param config: Config with the parameters specified in Config class
        :return: Tuple with the QUBO, time it took to map it
        """""
        start = start_time_measurement()

        # Extract hard and soft constraints from the generated problem
        hard, soft = problem

        # Count the variables
        self.nr_vars = len(hard.vars().union(And(soft).vars()))
        lagrange = config['lagrange']
        # Lagrange parameter is a factor of the number of soft constraints.
        lagrange *= len(soft)

        def _add_clause(curr_qubo_dict: dict[tuple[int, int], float],
                        clause: any,
                        pos: int,
                        weight: float) -> dict[tuple[int, int], float]:
            """
            Function that adds the QUBO terms corresponding to the clause and updates the QUBO dictionary
             accordingly. Additionally, the weight of the clause is taken into account.

            :param curr_qubo_dict: Current QUBO dictionary
            :param clause: Clause to be added
            :param pos: Position of the auxiliary variable
            :param weight: Weight of the clause
            :return: Updated QUBO dictionary
            """

            def _check_and_add(dictionary: dict, key: tuple[int, int], value: float) -> dict:
                """
                Helper function that checks if key is present or not in dictionary and adds a value, adding the key
                if missing.

                :param dictionary: Dictionary to be updated
                :param key: Key to check in the dictionary
                :param value: Value to add to the key
                :return: Updated dictionary
                """
                key = tuple(sorted(key))
                if key not in dictionary.keys():
                    dictionary[key] = value
                else:
                    dictionary[key] += value
                return dictionary

            cl_dict = {}
            for variable in clause.children:
                for variable_name in variable.vars():
                    # Transforms the negations (0,1) into signs (-1, 1)
                    cl_dict[int(variable_name[1:])] = (int(variable.true) - 1 / 2) * 2

            # Add the linear term of the auxiliary variable w
            curr_qubo_dict = _check_and_add(curr_qubo_dict, (pos, pos), 2 * weight)

            # Add x linear terms and xw terms.
            for qvar, val in cl_dict.items():
                # qvar is the name of the var, val is the sign corresponding to whether the variable is negated or not.
                # linear x term:
                curr_qubo_dict = _check_and_add(curr_qubo_dict, (qvar, qvar), -weight * val)
                # x * w (aux. var.) term
                curr_qubo_dict = _check_and_add(curr_qubo_dict, (qvar, pos), -weight * val)
            # Add combinations
            for q1, q2 in combinations(cl_dict.keys(), 2):
                curr_qubo_dict = _check_and_add(curr_qubo_dict, (q1, q2), weight * cl_dict[q1] * cl_dict[q2])

            return curr_qubo_dict

        qubo_dict = {}
        # Add the hard constraints and add the lagrange parameter as weight
        for clause_ind, hard_clause in enumerate(hard):
            qubo_dict = _add_clause(qubo_dict, hard_clause, self.nr_vars + clause_ind, lagrange)

        # Add the soft constraints and start the enumeration at the final index corresponding to hard cons.
        for clause_ind, soft_clause in enumerate(soft):
            qubo_dict = _add_clause(qubo_dict, soft_clause, self.nr_vars + clause_ind + len(hard), 1)

        logging.info(
            f"Generate Dinneen QUBO with {self.nr_vars + len(hard) + len(soft)} binary variables."
            f" Lagrange parameter used was: {config['lagrange']}."
        )

        return {"Q": qubo_dict}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Reverse mapping of the solution obtained from the Dinneen QUBO.

        :param solution: Dictionary containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()
        mapped_sol = {}
        for i in range(self.nr_vars):
            # if variable not present in solution, its assignment does not matter
            if i not in solution.keys():
                mapped_sol[f'L{i}'] = True
            else:
                mapped_sol[f'L{i}'] = bool(solution[i])

        return mapped_sol, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TypedDict

from qubovert.sat import NOT, OR, AND
from nnf import And

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class QubovertQUBO(Mapping):
    """
    Qubovert formulation of the vehicle-options problem.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]
        self.pubo_problem = None
        self.nr_vars = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "nnf", "version": "0.4.1"},
            {"name": "qubovert", "version": "1.2.5"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dict with configurable settings
        .. code-block:: python

            return {
                    "lagrange": {
                        "values": [0.1, 1, 1.5, 2, 5, 10, 1000, 10000],
                        "description": "By which factor would you like to multiply your Lagrange?"
                    }
                }
        """
        return {
            "lagrange": {
                "values": [0.1, 1, 1.5, 2, 5, 10, 1000, 10000],
                "description": "By which factor would you like to multiply your Lagrange?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             lagrange: float

        """
        lagrange: float

    @staticmethod
    def _constraints2qubovert(constraints: any) -> AND:
        """
        Converts the constraints nnf to a PUBO in the qubovert library.

        :param constraints: Constraints in nnf format
        :return: Constraints in qubovert format
        """
        clauses = []
        for c in constraints.children:
            literals = [v.name if v.true else NOT(v.name) for v in c.children]
            clauses.append(OR(*literals))
        return AND(*clauses)

    @staticmethod
    def _tests2qubovert(test_clauses: dict) -> sum:
        """
        Converts the list of test clauses in the nnf format to a PUBO.

        :param test_clauses: Test clauses in nnf format
        :return: Sum of mapped test clauses
        """
        mapped_tests = []

        for test_clause in test_clauses:
            mapped_tests.append(OR(*[v.name if v.true else NOT(v.name) for v in test_clause.children]))

        return sum(mapped_tests)

    def map(self, problem: any, config: Config) -> tuple[dict, float]:
        """
        Converts the problem to a QUBO in dictionary format. Problem is a CNF formula from the nnf library.

        :param problem: SAT problem
        :param config: Config with the parameters specified in Config class
        :return: Dict with the QUBO, time it took to map it
        """
        start = start_time_measurement()
        lagrange = config['lagrange']

        constraints, test_clauses = problem

        # Find number of the variables that appear in the tests and constraints, to verify the reverse mapping.
        self.nr_vars = len(constraints.vars().union(And(test_clauses).vars()))

        # Convert the constraints to qubovert:
        constraints_pubo = self._constraints2qubovert(constraints)

        # Convert the tests into qubovert:
        tests_pubo = self._tests2qubovert(test_clauses)
        logging.info(f'{tests_pubo.to_qubo().num_terms} number of terms in tests qubo')
        lagrange *= len(test_clauses)

        # Define the total PUBO problem:
        self.pubo_problem = -(tests_pubo + lagrange * constraints_pubo)

        # Convert to qubo:
        qubo_problem = self.pubo_problem.to_qubo()
        qubo_problem.normalize()
        logging.info(f"Converted to QUBO with {qubo_problem.num_binary_variables} Variables."
                     f" Lagrange parameter: {config['lagrange']}.")

        # Convert it to the right format to be accepted by Braket / Dwave
        q_dict = {}

        for k, v in qubo_problem.items():
            # "interaction (quadratic) terms":
            if len(k) == 2:
                if (k[0], k[1]) not in q_dict:
                    q_dict[(k[0], k[1])] = float(v)
                else:
                    q_dict[(k[0], k[1])] += float(v)
            # "local (linear) fields":
            if len(k) == 1:
                if (k[0], k[0]) not in q_dict:
                    q_dict[(k[0], k[0])] = float(v)
                else:
                    q_dict[(k[0], k[0])] += float(v)

        return {"Q": q_dict}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the SAT class for validation/evaluation.

        :param solution: Dictionary containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()
        pubo_sol = self.pubo_problem.convert_solution(solution)

        # Check if all variables appear in the solution.
        missing_vars = {f'L{i}' for i in range(self.nr_vars)} - set(pubo_sol.keys())

        # Add values for the missing variables -- if they do not appear, then their assignment does not matter.
        for missing_var in missing_vars:
            pubo_sol[missing_var] = True

        return pubo_sol, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Union
import numpy as np

from qiskit import QuantumCircuit, transpile
from qiskit.circuit import Parameter
from qiskit.providers import Backend
from qiskit.quantum_info import Statevector

from quark.modules.applications.qml.generative_modeling.training.QCBM import QCBM
from quark.modules.applications.qml.generative_modeling.training.QGAN import QGAN
from quark.modules.applications.qml.generative_modeling.training.Inference import Inference
from quark.modules.applications.qml.generative_modeling.mappings.LibraryGenerative import LibraryGenerative

logging.getLogger("qiskit").setLevel(logging.WARNING)


class LibraryQiskit(LibraryGenerative):
    """
    This module maps a library-agnostic gate sequence to a qiskit circuit.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("LibraryQiskit")
        self.submodule_options = ["QCBM", "QGAN", "Inference"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "qiskit", "version": "1.3.0"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for the Qiskit Library.

        :return: Dictionary with configurable settings
        .. code-block:: python

        return {
            "backend": {
                "values": ["aer_statevector_simulator_gpu", "aer_statevector_simulator_cpu",
                           "cusvaer_simulator (only available in cuQuantum appliance)", "aer_simulator_gpu",
                           "aer_simulator_cpu", "ionQ_Harmony", "Amazon_SV1",
                           "simulator_statevector IBM Quantum Platform", "ibm_brisbane IBM Quantum Platform"],
                "description": "Which backend do you want to use? (aer_statevector_simulator\
                             uses the measurement probability vector, the others are shot based)"
            },

            "n_shots": {
                "values": [100, 1000, 10000, 1000000],
                "description": "How many shots do you want use for estimating the PMF of the model?\
                 (If the aer_statevector_simulator selected, only relevant for studying generalization)"
            }
        }
        """

        return {
            "backend": {
                "values": ["aer_statevector_simulator_gpu", "aer_statevector_simulator_cpu",
                           "cusvaer_simulator (only available in cuQuantum appliance)", "aer_simulator_gpu",
                           "aer_simulator_cpu", "ionQ_Harmony", "Amazon_SV1", "ibm_brisbane IBM Quantum Platform"],
                "description": "Which backend do you want to use? (aer_statevector_simulator uses the measurement "
                               "probability vector, the others are shot based)"
            },
            "n_shots": {
                "values": [100, 1000, 10000, 1000000],
                "description": "How many shots do you want use for estimating the PMF of the model? "
                               "(If the aer_statevector_simulator selected, only relevant for studying generalization)"
            }
        }

    def get_default_submodule(self, option: str) -> Union[QCBM, QGAN, Inference]:
        """
        Returns the default submodule based on the provided option.

        :param option: The option to select the submodule
        :return: The selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "QCBM":
            return QCBM()
        elif option == "QGAN":
            return QGAN()
        elif option == "Inference":
            return Inference()
        else:
            raise NotImplementedError(f"Option {option} not implemented")

    def sequence_to_circuit(self, input_data: dict) -> dict:
        """
        Maps the gate sequence, that specifies the architecture of a quantum circuit
        to its Qiskit implementation.

        :param input_data: Collected information of the benchmarking process
        :return: Same dictionary but the gate sequence is replaced by its Qiskit implementation
        """
        n_qubits = input_data["n_qubits"]
        gate_sequence = input_data["gate_sequence"]

        circuit = QuantumCircuit(n_qubits, n_qubits)
        param_counter = 0
        for gate, wires in gate_sequence:
            if gate == "Hadamard":
                circuit.h(wires[0])
            elif gate == "CNOT":
                circuit.cx(wires[0], wires[1])
            elif gate == "RZ":
                circuit.rz(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RX":
                circuit.rx(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RY":
                circuit.ry(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RXX":
                circuit.rxx(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "RYY":
                circuit.ryy(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "RZZ":
                circuit.rzz(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "CRY":
                circuit.cry(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "Barrier":
                circuit.barrier()
            elif gate == "Measure":
                circuit.measure(wires[0], wires[0])
            else:
                raise NotImplementedError(f"Gate {gate} not implemented")

        input_data["circuit"] = circuit
        input_data.pop("gate_sequence")
        input_data["n_params"] = len(circuit.parameters)

        return input_data

    @staticmethod
    def select_backend(config: str, n_qubits: int) -> any:
        """
        This method configures the backend.

        :param config: Name of a backend
        :param n_qubits: Number of qubits
        :return: Configured qiskit backend
        """
        if config == "cusvaer_simulator (only available in cuQuantum appliance)":
            import cusvaer  # pylint: disable=C0415
            from qiskit_aer import AerSimulator  # pylint: disable=C0415
            backend = AerSimulator(
                method="statevector",
                device="GPU",
                cusvaer_enable=True,
                noise_model=None,
                cusvaer_p2p_device_bits=3,
                cusvaer_comm_plugin_type=cusvaer.CommPluginType.MPI_AUTO,
                cusvaer_comm_plugin_soname="libmpi.so"
            )

        elif config == "aer_simulator_gpu":
            from qiskit_aer import Aer  # pylint: disable=C0415
            backend = Aer.get_backend("aer_simulator")
            backend.set_options(device="GPU")
        elif config == "aer_simulator_cpu":
            from qiskit_aer import Aer  # pylint: disable=C0415
            backend = Aer.get_backend("aer_simulator")
            backend.set_options(device="CPU")
        elif config == "aer_statevector_simulator_gpu":
            from qiskit_aer import Aer  # pylint: disable=C0415
            backend = Aer.get_backend('statevector_simulator')
            backend.set_options(device="GPU")
        elif config == "aer_statevector_simulator_cpu":
            from qiskit_aer import Aer  # pylint: disable=C0415
            backend = Aer.get_backend('statevector_simulator')
            backend.set_options(device="CPU")
        elif config == "ionQ_Harmony":
            from quark.modules.devices.braket.Ionq import Ionq  # pylint: disable=C0415
            from qiskit_braket_provider import AWSBraketBackend, AWSBraketProvider  # pylint: disable=C0415
            device_wrapper = Ionq("ionQ", "arn:aws:braket:::device/qpu/ionq/ionQdevice")
            backend = AWSBraketBackend(
                device=device_wrapper.device,
                provider=AWSBraketProvider(),
                name=device_wrapper.device.name,
                description=f"AWS Device: {device_wrapper.device.provider_name} {device_wrapper.device.name}.",
                online_date=device_wrapper.device.properties.service.updatedAt,
                backend_version="2",
            )
        elif config == "Amazon_SV1":
            from quark.modules.devices.braket.SV1 import SV1  # pylint: disable=C0415
            from qiskit_braket_provider import AWSBraketBackend, AWSBraketProvider  # pylint: disable=C0415
            device_wrapper = SV1("SV1", "arn:aws:braket:::device/quantum-simulator/amazon/sv1")
            backend = AWSBraketBackend(
                device=device_wrapper.device,
                provider=AWSBraketProvider(),
                name=device_wrapper.device.name,
                description=f"AWS Device: {device_wrapper.device.provider_name} {device_wrapper.device.name}.",
                online_date=device_wrapper.device.properties.service.updatedAt,
                backend_version="2",
            )

        else:
            raise NotImplementedError(f"Device Configuration {config} not implemented")

        return backend

    @staticmethod
    def get_execute_circuit(circuit: QuantumCircuit, backend: Backend, config: str, config_dict: dict) \
            -> tuple[any, any]:  # pylint: disable=W0221,R0915
        """
        This method combines the qiskit circuit implementation and the selected backend and returns a function,
        that will be called during training.

        :param circuit: Qiskit implementation of the quantum circuit
        :param backend: Configured qiskit backend
        :param config: Name of a backend
        :param config_dict: Contains information about config
        :return: Tuple that contains a method that executes the quantum circuit for a given set of parameters and the
        transpiled circuit
        """
        n_shots = config_dict["n_shots"]
        n_qubits = circuit.num_qubits
        circuit_transpiled = transpile(circuit, backend=backend)

        if config in ["aer_statevector_simulator_gpu", "aer_statevector_simulator_cpu"]:
            circuit_transpiled.remove_final_measurements()

            def execute_circuit(solutions):
                all_circuits = [
                    circuit_transpiled.assign_parameters(solution) for solution in solutions
                ]
                pmfs = np.asarray([Statevector(c).probabilities() for c in all_circuits])
                return pmfs, None

        elif config in ["ionQ_Harmony", "Amazon_SV1"]:
            import time as timetest  # pylint: disable=C0415

            def execute_circuit(solutions):
                all_circuits = [
                    circuit_transpiled.assign_parameters(solution) for solution in solutions
                ]
                jobs = backend.run(all_circuits, shots=n_shots)
                while not jobs.in_final_state():
                    logging.info("Waiting 10 seconds for task to finish")
                    timetest.sleep(10)

                samples_dictionary = [
                    jobs.result().get_counts(circuit).int_outcomes() for circuit in all_circuits
                ]

                samples = []
                for result in samples_dictionary:
                    target_iter = np.zeros(2 ** n_qubits)
                    result_keys = list(result.keys())
                    result_vals = list(result.values())
                    target_iter[result_keys] = result_vals
                    target_iter = np.asarray(target_iter)
                    samples.append(target_iter)

                samples = np.asarray(samples)
                pmfs = samples / n_shots

                return pmfs, samples

        elif config in [
            "cusvaer_simulator (only available in cuQuantum appliance)",
            "aer_simulator_cpu",
            "aer_simulator_gpu"
        ]:

            def execute_circuit(solutions):
                all_circuits = [
                    circuit_transpiled.assign_parameters(solution) for solution in solutions
                ]
                jobs = backend.run(all_circuits, shots=n_shots)
                samples_dictionary = [
                    jobs.result().get_counts(circuit).int_outcomes() for circuit in all_circuits
                ]

                samples = []
                for result in samples_dictionary:
                    target_iter = np.zeros(2 ** n_qubits)
                    result_keys = list(result.keys())
                    result_vals = list(result.values())
                    target_iter[result_keys] = result_vals
                    target_iter = np.asarray(target_iter)
                    samples.append(target_iter)

                samples = np.asarray(samples)
                pmfs = samples / n_shots
                return pmfs, samples

        else:
            logging.error(f"Unknown backend option selected: {config}")
            logging.error("Run terminates with Exception error.")
            raise Exception

        return execute_circuit, circuit_transpiled



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import getpass
import logging
import os
from abc import ABC
from datetime import datetime

import boto3
from botocore.config import Config
from botocore.exceptions import ProfileNotFound
from braket.aws import AwsSession

from quark.modules.devices.Device import Device


class Braket(Device, ABC):
    """
    Abstract class to use the Amazon Braket devices.
    """

    def __init__(self, device_name: str, region: str = None, arn: str = None):
        """
        Constructor method.
        """
        super().__init__(device_name)
        self.device = None
        self.arn = arn
        self.s3_destination_folder = None
        self.boto_session = None
        self.aws_session = None

        if 'SKIP_INIT' in os.environ:
            # TODO: This is currently needed so create_module_db in the Installer does not execute the rest
            #       of this section, which would be unnecessary. However, this should be done better in the future!
            return

        if device_name != "LocalSimulator":
            self._configure_aws_session(region)

    def _configure_aws_session(self, region: str) -> None:
        """
        Configures the AWS session for the Braket device.

        :param region: AWS region to use
        """
        proxy_definitions = self._setup_proxy()
        region = self._set_region(region)
        my_config = Config(region_name=region, proxies=proxy_definitions)

        profile_name = self._set_profile()
        self._initialize_aws_session(profile_name, region, my_config)

    @staticmethod
    def _setup_proxy() -> any:
        """
        Sets up proxy configuration if available in the environment variables.

        :return: Proxy definitions
        """
        if 'HTTP_PROXY' in os.environ:
            proxy_definitions = {
                'http': os.environ['HTTP_PROXY'],
                'https': os.environ['HTTP_PROXY']
            }
            os.environ['HTTPS_PROXY'] = os.environ['HTTP_PROXY']
        else:
            logging.warning('No HTTP_PROXY set as an environment variable. '
                            'This might cause trouble if you are using a VPN.')
            proxy_definitions = None
        return proxy_definitions

    @staticmethod
    def _set_region(region: str) -> str:
        """
        Sets the AWS region from the environment variable or defaults to 'us-east-1'.

        :param region: Provided region
        :return: Final region to be used
        """
        if region is None:
            region = os.environ.get('AWS_REGION', 'us-east-1')
            logging.info(f"No AWS_REGION specified, using default region: {region}")
        return region

    @staticmethod
    def _set_profile() -> str:
        """
        Determines the AWS profile to use for the session.

        :return: AWS profile name
        """
        if 'AWS_PROFILE' in os.environ:
            return os.environ['AWS_PROFILE']
        elif "AWS_CONTAINER_CREDENTIALS_RELATIVE_URI" in os.environ:
            logging.info("Assuming AWS container environment, using container credentials.")
            return None
        else:
            profile_name = 'quantum_computing'
            os.environ['AWS_PROFILE'] = profile_name
            logging.info(f"No AWS_PROFILE specified, using default profile: {profile_name}")
            return profile_name

    def _initialize_aws_session(self, profile_name: str, region: str, my_config: Config) -> None:
        """
        Initializes the AWS session for interacting with Amazon Braket.

        :param profile_name: AWS profile name
        :param region: AWS region
        :param my_config: Boto3 configuration
        :raises Exception: If the AWS profile is not found
        """
        try:
            if profile_name is None:
                self.boto_session = boto3.Session(region_name=region)
            else:
                self.boto_session = boto3.Session(profile_name=profile_name, region_name=region)
            self.aws_session = AwsSession(boto_session=self.boto_session, config=my_config)
        except ProfileNotFound as exc:
            logging.error(f"AWS-Profile {profile_name} could not be found! Please set the AWS_PROFILE env variable. "
                          "Only LocalSimulator is available.")
            raise Exception("Please refer to the logged error message.") from exc

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dictionaries with requirements
        """
        return [
            {"name": "amazon-braket-sdk", "version": "1.88.2"},
            {"name": "botocore", "version": "1.35.73"},
            {"name": "boto3", "version": "1.35.73"}
        ]

    def init_s3_storage(self, folder_name: str) -> None:
        """
        Initializes an S3 storage bucket for Amazon Braket.

        :param folder_name: Name of the s3 bucket
        """
        run_timestamp = datetime.today().date()
        username = getpass.getuser()

        bucket_name = f"amazon-braket-benchmark-framework-{run_timestamp}-{username}"
        self.s3_destination_folder = (bucket_name, folder_name)
        self._create_s3_bucket(self.boto_session, bucket_name)

    @staticmethod
    def _create_s3_bucket(boto3_session: boto3.Session, bucket_name: str = 'quark-benchmark-framework',
                          region: str = 'us-east-1') -> None:
        """
        Creates an S3 bucket with specific configurations.

        :param boto3-session: Boto3 session
        :param bucket_name: Name of the s3 bucket
        :param region: AWS region
        """
        s3_client = boto3_session.client('s3', region_name=region)

        if region == "us-east-1":
            s3_client.create_bucket(Bucket=bucket_name)
        else:
            location = {"LocationConstraint": region}
            s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration=location)

        s3_client.put_public_access_block(
            Bucket=bucket_name,
            PublicAccessBlockConfiguration={
                'BlockPublicAcls': True,
                'IgnorePublicAcls': True,
                'BlockPublicPolicy': True,
                'RestrictPublicBuckets': True
            },
        )



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union
from quark.utils import start_time_measurement, end_time_measurement

from quark.modules.applications.qml.QML import QML
from quark.modules.applications.qml.generative_modeling.data.data_handler.DiscreteData import DiscreteData
from quark.modules.applications.qml.generative_modeling.data.data_handler.ContinuousData import ContinuousData


class GenerativeModeling(QML):
    """
    Generative models enable the creation of new data by learning the underlying probability distribution of the
    training data set of interest. More specifically, a generative model attempts to learn an unknown probability
    distribution Q by modeling an approximated probability distribution P(θ) which is parameterized by a set of
    variables θ. Data sampled from Q is used to train a model by tuning θ such that P(θ) more closely approximates Q.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("GenerativeModeling")
        self.submodule_options = ["Continuous Data", "Discrete Data"]
        self.data = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dicts with requirements of this module
        """
        return []

    def get_solution_quality_unit(self) -> str:
        return "minimum KL"

    def get_default_submodule(self, option: str) -> Union[ContinuousData, DiscreteData]:
        """
        Returns the default submodule based on the given option.

        :param option: The submodule option to select
        :return: Instance of the selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "Continuous Data":
            self.data = ContinuousData()
        elif option == "Discrete Data":
            self.data = DiscreteData()
        else:
            raise NotImplementedError(f"Transformation Option {option} not implemented")
        return self.data

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: Dictionary of configurable parameters
        .. code-block:: python

            return {
                    "n_qubits": {
                        "values": [4, 6, 8, 10, 12],
                        "description": "How many qubits do you want to use?"
                    }
                }
        """
        return {
            "n_qubits": {
                "values": [6, 8, 10, 12],
                "description": "How many qubits do you want to use?"
            }
        }

    def generate_problem(self, config: dict) -> dict:
        """
        The number of qubits is chosen for this problem.

        :param config: Dictionary including the number of qubits
        :return: Dictionary with the number of qubits
        """
        application_config = {"n_qubits": config["n_qubits"]}
        return application_config

    def preprocess(self, input_data: dict, config: dict, **kwargs: dict) -> tuple[dict, float]:
        """
        Generate the actual problem instance in the preprocess function.

        :param input_data: Usually not used for this method
        :param config: Config for the problem creation
        :param kwargs: Optional additional arguments
        :return: Tuple containing qubit number and the function's computation time
        """
        start = start_time_measurement()
        output = self.generate_problem(config)
        output["store_dir_iter"] = f"{kwargs['store_dir']}/rep_{kwargs['rep_count']}"
        return output, end_time_measurement(start)

    def postprocess(self, input_data: dict, config: dict, **kwargs: dict) -> tuple[dict, float]:
        """
        Process the solution here, then validate and evaluate it.

        :param input_data: A representation of the quantum machine learning model that will be trained
        :param config: Config specifying the parameters of the training
        :param kwargs: Optional keyword arguments
        :return: Tuple with input_data and the function's computation time
        """
        start = start_time_measurement()
        return input_data, end_time_measurement(start)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import networkx as nx
import pickle

# Create the original graph as a MultiDiGraph
graph = nx.MultiDiGraph()

with open("reference_data.txt") as infile:
    for line in infile:
        line_elements = line.split()

        # Extract start and end attributes from line elements
        r_start, s_start, n_start, c_start, t_start, l_start = map(int, line_elements[1:7])
        r_end, s_end, n_end, c_end, t_end, l_end = map(int, line_elements[8:14])
        duration = float(line_elements[15])

        # Handle missing or invalid data with default values
        if s_start == -1:
            s_start = 0
            t_start = 1  # TODO except of picking a hardcoded value here we should select 1 from the dataset itself
            c_start = 1
        if s_end == -1:
            s_end = 0
            t_end = 1
            c_end = 1
        if n_start == -1:
            n_start = 0
        if n_end == -1:
            n_end = 0

        # Reduce the number of tools and configurations for simplicity
        if c_end < 3 and c_start < 3 and t_start < 2 and t_end < 2:
            graph.add_edge(
                (s_start, n_start), (s_end, n_end),
                c_start=c_start, t_start=t_start,
                c_end=c_end, t_end=t_end, weight=duration
            )

# Save the graph to a file in gpickle format
with open("reference_graph.gpickle", "wb") as file:
    pickle.dump(graph, file, pickle.HIGHEST_PROTOCOL)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement

from quark.modules.applications.qml.Circuit import Circuit


class CircuitGenerative(Circuit, Core, ABC):
    """
    This module is abstract base class for the library-agnostic gate sequence, that define a quantum circuit.
    """

    def __init__(self, name: str):
        """
        Constructor method.

        :param name: The name of the circuit architecture
        """
        super().__init__()
        self.architecture_name = name

    def preprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[dict, float]:
        """
        Library-agnostic implementation of the gate sequence, that will be mapped to backend such as Qiskit in the
         subsequent module.

        :param input_data: Collection of information from the previous modules
        :param config: Config specifying the number of qubits of the circuit
        :param kwargs: Optional keyword arguments
        :return: Dictionary including the dataset, the gate sequence needed for circuit construction, and the time it
                 took generate the gate sequence.
        """
        start = start_time_measurement()
        circuit_constr = self.generate_gate_sequence(input_data, config)

        if "generalization_metrics" in input_data:
            circuit_constr["generalization_metrics"] = input_data["generalization_metrics"]

        return circuit_constr, end_time_measurement(start)

    def postprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[dict, float]:
        """
        Method that passes back information of the subsequent modules to the preceding modules.

        :param input_data: Collected information of the benchmarking process
        :param config: Config specifying the number of qubits of the circuit
        :param kwargs: Optional keyword arguments
        :return: Same dictionary like input_data with architecture_name and execution time
        """
        start = start_time_measurement()
        input_data["architecture_name"] = self.architecture_name
        return input_data, end_time_measurement(start)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and

from abc import ABC, abstractmethod


class Model(ABC):
    """
    Abstract base class for any quantum model. This class defines the necessary methods
    that models like 'LibraryGenerative' must implement.
    """

    @abstractmethod
    def sequence_to_circuit(self, input_data: dict) -> dict:
        """
        Abstract method to convert a sequence into a quantum circuit.

        :param input_data: Input data representing the gate sequence
        :return: A dictionary representing the quantum circuit
        """
        pass

    @staticmethod
    @abstractmethod
    def get_execute_circuit(circuit: any, backend: any, config: str, config_dict: dict) -> tuple[any, any]:
        """
        This method combines the circuit implementation and the selected backend and returns a function that will be
        called during training.

        :param circuit: Implementation of the quantum circuit
        :param backend: Configured qiskit backend
        :param config: Name of a backend
        :param config_dict: Dictionary including the number of shots
        :return: Tuple that contains a method that executes the quantum circuit for a given set of parameters and the
        transpiled circuit
        """
        pass

    @staticmethod
    @abstractmethod
    def select_backend(config: str, n_qubits: int) -> any:
        """
        This method configures the backend.

        :param config: Name of a backend
        :param n_qubits: Number of qubits
        :return: Configured backend
        """
        pass
